/******************************************************************************
*******************************************************************************
*******************************************************************************

    redland RDF private support functions.  

    Copyright (C) 2003 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: Redland_private.hh,v 1.14 2007/12/14 21:30:10 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#ifndef _ALREADY_INCLUDED_FERRIS_REDLAND_PRIV_H_
#define _ALREADY_INCLUDED_FERRIS_REDLAND_PRIV_H_

#include <Ferris/HiddenSymbolSupport.hh>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <redland.h>

#include <Ferris/TypeDecl.hh>
#include <FerrisStreams/All.hh>

#include <iterator>

namespace boost {
    namespace serialization {
        class access;
    }
}

namespace Ferris
{
    namespace RDF 
    {
        extern const std::string RDF_FERRIS_BASE;

        
        class World;
        class URI;
        class Storage;
        class Parser;
        class Model;
        class Statement;
        class Node;
        class RedlandStreamIterator;
        class RedlandIterator;
        class StatementIterator;
        class ObjectIterator;
        class BindingsIterator;
        FERRIS_SMARTPTR( World,     fh_world );
        FERRIS_SMARTPTR( URI,       fh_uri );
        FERRIS_SMARTPTR( Storage,   fh_storage );
        FERRIS_SMARTPTR( Parser,    fh_parser );
        FERRIS_SMARTPTR( Model,     fh_model );
        FERRIS_SMARTPTR( Statement, fh_statement );
        FERRIS_SMARTPTR( Node,      fh_node );

        /**
         * A list of statements
         */
        typedef std::list< fh_statement > StatementList_t;
        
        class FERRISEXP_API World
            :
            public Handlable
        {
            librdf_world* m_world;

            NOT_COPYABLE( World );
            
        public:
            
            World();
            virtual ~World();
            librdf_world* getRAW();
        };
        
        FERRISEXP_API fh_world getWorld();
        FERRISEXP_API fh_model getDefaultFerrisModel();

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        FERRISEXP_API bool operator==(const RedlandStreamIterator& __x, const RedlandStreamIterator& __y);
        FERRISEXP_API bool operator!=(const RedlandStreamIterator& __x, const RedlandStreamIterator& __y);
        class FERRISEXP_API RedlandStreamIterator
        {
            /**
             * Simple class to allow intrusive reference counting while at the
             * same time avoiding a custom storage class "RedlandStreamSPStorage"
             * which knows how to deallocate a librdf_stream object.
             */
            struct librdf_stream_holder
                :
                public Handlable
            {
            private:
                NOT_COPYABLE( librdf_stream_holder );
            public:
                librdf_stream* m_stream;
                librdf_stream_holder( librdf_stream* s = 0 )
                    :
                    m_stream( s )
                    {}
                virtual ~librdf_stream_holder()
                    {
                        if( m_stream )
                            librdf_free_stream( m_stream );
                    }
            };
            
            typedef RedlandStreamIterator _Self;
            friend class Model;

            void __free();
            void __advance();
            
        protected:
            librdf_stream* __stream();
            FERRIS_SMARTPTR( librdf_stream_holder, fh_stream );

            fh_stream   m_stream;
            
            explicit RedlandStreamIterator( librdf_stream* stream );

        public:

            typedef std::forward_iterator_tag iterator_category;
            typedef long difference_type;

            RedlandStreamIterator();
            ~RedlandStreamIterator();

            _Self& operator++() {
                __advance();
                return *this;
            }
            _Self operator++(int) {
                _Self __tmp = *this;
                __advance();
                return __tmp;
            }
            friend bool operator==(const RedlandStreamIterator& __x, const RedlandStreamIterator& __y);
        };


        class FERRISEXP_API StatementIterator
            :
            public RedlandStreamIterator
        {
            typedef StatementIterator _Self;
            friend class Model;

        protected:
            
            explicit StatementIterator( librdf_stream* stream )
                : RedlandStreamIterator( stream )
                {}
            
        public:

            typedef std::forward_iterator_tag iterator_category;
            typedef fh_statement value_type;
            typedef long         difference_type;
            typedef fh_statement pointer;
            typedef fh_statement reference;

            StatementIterator()     {}
            ~StatementIterator()    {}
            reference operator*();
        };
        

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        FERRISEXP_API bool operator==(const RedlandIterator& __x, const RedlandIterator& __y);
        FERRISEXP_API bool operator!=(const RedlandIterator& __x, const RedlandIterator& __y);
        class FERRISEXP_API RedlandIterator
        {
            typedef RedlandIterator _Self;
            friend class Model;

            /**
             * Simple class to allow intrusive reference counting while at the
             * same time avoiding a custom storage class "RedlandIteratorSPStorage"
             * which knows how to deallocate a librdf_iterator object.
             */
            struct librdf_iterator_holder
                :
                public Handlable
            {
            private:
                NOT_COPYABLE( librdf_iterator_holder );
            public:
                librdf_iterator* m_iter;
                librdf_iterator_holder( librdf_iterator* i = 0 );
                virtual ~librdf_iterator_holder();
            };
            
            
            void __free();
            void __advance();
            
        protected:

            librdf_iterator* __iter();
            FERRIS_SMARTPTR( librdf_iterator_holder, fh_iterator );
            fh_iterator   m_iter;
            
            explicit RedlandIterator( librdf_iterator* iter );

        public:

            typedef std::forward_iterator_tag iterator_category;
            typedef long difference_type;

            RedlandIterator();
            ~RedlandIterator();

            _Self& operator++() {
                __advance();
                return *this;
            }
            _Self operator++(int) {
                _Self __tmp = *this;
                __advance();
                return __tmp;
            }
            friend bool operator==(const RedlandIterator& __x, const RedlandIterator& __y);
        };


        class FERRISEXP_API ObjectIterator
            :
            public RedlandIterator
        {
            typedef ObjectIterator _Self;
            friend class Model;

            explicit ObjectIterator( librdf_iterator* iter );
            
        public:

            typedef std::forward_iterator_tag iterator_category;
            typedef fh_node value_type;
            typedef long    difference_type;
            typedef fh_node pointer;
            typedef fh_node reference;

            ObjectIterator();
            ~ObjectIterator();
            reference operator*();
        };
        

        FERRISEXP_API bool operator==(const BindingsIterator& __x, const BindingsIterator& __y);
        FERRISEXP_API bool operator!=(const BindingsIterator& __x, const BindingsIterator& __y);
        class FERRISEXP_API BindingsIterator
        {
            typedef BindingsIterator _Self;
            friend class Model;

        public:

            class FERRISEXP_API binding 
            {
            public:
                typedef std::vector< std::string > m_bindingNames_t;

                int size() const;
                const m_bindingNames_t& getNames() const;
                fh_node getByName( const std::string& n ) const;
                fh_node getByID( int id ) const;
                std::string getName( int id ) const;
                
            private:
                friend class BindingsIterator;
                m_bindingNames_t m_bindingNames;
                
                typedef std::vector< fh_node > d_t;
                d_t d;
                void setup( librdf_query_results* r );
            };
        private:
            binding m_binding;
            
            /**
             * Simple class to allow intrusive reference counting while at the
             * same time avoiding a custom storage class "RedlandIteratorSPStorage"
             * which knows how to deallocate a librdf_iterator object.
             */
            struct librdf_query_results_holder
                :
                public Handlable
            {
            private:
                NOT_COPYABLE( librdf_query_results_holder );
            public:
                librdf_query_results* m_iter;
                librdf_query_results_holder( librdf_query_results* i = 0 )
                    :
                    m_iter( i )
                    {}
                virtual ~librdf_query_results_holder()
                    {
                        if( m_iter )
                            librdf_free_query_results( m_iter );
                    }
            };
            
            explicit BindingsIterator( librdf_query_results* iter );
            void __free();
            void __advance();
            
        protected:

            librdf_query_results* __iter();
            FERRIS_SMARTPTR( librdf_query_results_holder, fh_iterator );
            fh_iterator   m_iter;
            
        public:

            typedef std::forward_iterator_tag iterator_category;
            typedef binding value_type;
            typedef long    difference_type;
            typedef binding pointer;
            typedef binding& reference;

            BindingsIterator() : m_iter(0)    {}
            ~BindingsIterator()    {}

            reference operator*();
            _Self& operator++() {
                __advance();
                return *this;
            }
            _Self operator++(int) {
                _Self __tmp = *this;
                __advance();
                return __tmp;
            }
            friend bool operator==(const BindingsIterator& __x, const BindingsIterator& __y);
            
        };
        

        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        /******************************/
        /******************************/
        /******************************/

        class FERRISEXP_API Storage
            :
            public Handlable
        {
            librdf_storage* m_storage;
            
            Storage( const std::string& storage_name,
                     const std::string& name,
                     const std::string& options_string );
        public:
            static fh_storage CreateStorage( const std::string& storage_name = "memory",
                                             const std::string& name = "",
                                             const std::string& options_string = "" );
            static fh_storage CreateDBStorage( const std::string& name = "",
                                               const std::string& dir  = ".",
                                               bool writable = true );
            
            virtual ~Storage();
            librdf_storage* getRAW();
            
            fh_model CreateModel( const std::string& exopts = "" );
        };

        /******************************/
        /******************************/
        /******************************/

        class FERRISEXP_API Model
            :
            public Handlable
        {
            friend class    Storage;

            fh_storage      m_storage;
            librdf_model*   m_model;


            // This is a little bit hacky, basically, when we Load() a RDF/XML file
            // then we store its url so that we can stream it out when the model
            // object is sync()ed
            std::string m_rdfxml_url;

            Model( fh_storage s, const std::string& exopts = "" );
            Model( fh_storage s, librdf_model* m );

        public:

            /**
             * load an existing RDF store from 'c'
             */
            static fh_model Load( fh_context c );

            /**
             * load an existing RDF store from the string provided.
             */
            static fh_model Load( const std::string& rdfxml );

            /**
             * Merge some more RDF/XML with existing model
             */
            void MergeRDFXML( const std::string& rdfxml, fh_uri base_uri = 0 );
            
            /**
             * load or create a RDF store at 'c'
             */
            static fh_model ObtainDB( fh_context c,
                                      const std::string& name,
                                      const std::string& dbenv_name,
                                      bool writable = true,
                                      bool indexContexts = true,
                                      bool indexPredicates = true,
                                      bool createAsNew = false );
            static fh_model ObtainDB( fh_context c,
                                      const char* name,
                                      const char* dbenv_name,
                                      bool writable = true,
                                      bool indexContexts = true,
                                      bool indexPredicates = true,
                                      bool createAsNew = false );
            static fh_model ObtainDB( fh_context c,
                                      const std::string& name = "",
                                      bool writable = true,
                                      bool indexContexts = true,
                                      bool indexPredicates = true,
                                      bool createAsNew = false );
            static fh_model ObtainDB( const std::string& path,
                                      const std::string& name = "",
                                      bool writable = true,
                                      bool indexContexts = true,
                                      bool indexPredicates = true,
                                      bool createAsNew = false );
            static fh_model ObtainDB( const std::string& storage_name,
                                      const std::string& db_name,
                                      const std::string& db_options );
            static fh_model MemoryOnlyModel();
            
            virtual ~Model();
            librdf_model* getRAW() const;
            fh_model      clone();
            void          sync();

            int           size();
            void          insert( fh_statement st );
            void          insert( StatementIterator begin,
                                  StatementIterator end = StatementIterator() );
            void          insert( fh_node subj, fh_node pred, fh_node obj );
            void          insert( fh_model m );
            StatementIterator begin();
            StatementIterator end();
            
            void       addStatement( fh_statement st );
            fh_ostream write( fh_ostream oss,
                              const std::string& serializerName = "",
                              const std::string& mimeType = "",
                              fh_uri type_uri = 0 );

            StatementIterator find( const std::string& query,
                                    const std::string& queryTypeName = "",
                                    fh_uri queryTypeURI = 0 );
            BindingsIterator findBindings( const std::string& query,
                                           const std::string& queryTypeName = "sparql",
                                           fh_uri queryTypeURI = 0 );
            
            StatementIterator findStatements( fh_statement partial_statement );
            ObjectIterator    findSubjects(   fh_node pred, fh_node obj );
            ObjectIterator    findPredicates( fh_node subj, fh_node obj );
            ObjectIterator    findObjects(    fh_node subj, fh_node pred );
            fh_node           getSubject(     fh_node pred, fh_node obj );
            fh_node           getPredicate(   fh_node subj, fh_node obj );
            fh_node           getObject(      fh_node subj, fh_node pred );
            ObjectIterator    getArcsIn(  fh_node n );
            ObjectIterator    getArcsOut( fh_node n );
            bool              hasArcIn(   fh_node n, fh_node p );
            bool              hasArcOut(  fh_node n, fh_node p );
            
            bool contains( fh_statement st );
            void erase( fh_statement st );
            void erase( fh_model m );
            void erase( fh_node subj, fh_node pred, fh_node obj );
            void erase( fh_node subj, fh_node pred );
            void erase( fh_node subj, fh_node pred, ObjectIterator objiter );
            /**
             * Get rid of all triples such as
             * subj _p1 _o1
             * and any triples with _o1 as subject, and any of those
             * where the object is the subject of another triple.
             */
            void eraseTransitive( fh_node subj );

            /**
             * if there is a list/value for subj+pred remove it
             * set the triple subj, pred, obj after this.
             */
            void set( fh_node subj, fh_node pred, fh_node obj );

            /**
             * Get a tree with all the triples reachable from a single subject.
             */
            StatementList_t& getTransitiveTree( fh_node subj, StatementList_t& ret );


            /**
             * Any triples (?s,?p,?o) starting with source=?s are copied to target=?t
             * (?t,?p,?o) only when (?t,?p,???) does not exist.
             *
             * This is not transative
             */
            void RDFMerge( fh_node source, fh_node target );
            
        };

        /******************************/
        /******************************/
        /******************************/

        class FERRISEXP_API Parser
            :
            public Handlable
        {
            librdf_parser*  m_parser;

        public:
            Parser( const std::string& parser_name = "",
                    const std::string& mime_type = "",
                    fh_uri type_uri = 0 );
            
            virtual ~Parser();
            librdf_parser* getRAW();
            
            fh_model ParseIntoModel( fh_model m, fh_uri input_uri,     fh_uri base_uri = 0 );
            fh_model ParseIntoModel( fh_model m, const std::string& s, fh_uri base_uri = 0 );
            fh_model ParseIntoModel( fh_model m, fh_context c,         fh_uri base_uri = 0 );
            fh_model ParseIntoModel( fh_model m, fh_istream iss,       fh_uri base_uri = 0 );

            void setFeature( fh_uri feature, const std::string& s );
            void clearFeature( fh_uri feature );
            std::string getFeature( fh_uri feature );
        };
        
        
        
        /******************************/
        /******************************/
        /******************************/

        FERRISEXP_API bool operator==(const fh_node& __x, const fh_node& __y);
        FERRISEXP_API bool operator==(const Node& __x, const Node& __y);
        FERRISEXP_API bool operator!=(const Node& __x, const Node& __y);
        class FERRISEXP_API Node
            :
            public Handlable
        {
            friend class       ImplWorld;
            friend class       ObjectIterator;
            friend class       BindingsIterator::binding;
            
            librdf_node*       m_node;
            bool               m_ownNode;
            
            Node( librdf_node* n, bool takeOwnership = true );
            friend bool operator==(const Node& __x, const Node& __y);

        public:

            static fh_node CreateURI( const std::string& uri );
            static fh_node CreateLiteral( const std::string& s );
            static fh_node CreateLiteral( const std::wstring& s );
            static fh_node CreateBlankNode( const std::string& s = "" );

            
            virtual ~Node();
            librdf_node* getRAW() const;

            fh_uri getURI();
            librdf_node_type getType();

            std::string getLiteralValue();
            std::string getLiteralLang();
            fh_uri      getLiteralDatatype();
            int         getLIOrdinal();
            std::string getBlankID();

            bool isResource();
            bool isLiteral();
            bool isLiteralXML();
            bool isBlank();

            std::string toString();
//            std::string getDigest();

            std::string& encode( std::string& ret );
//            std::string& decode( const std::string& s );
        };
//         template < class _CharT, class _Traits >
//         std::basic_ostream< _CharT, _Traits >&
//         operator<<( std::basic_ostream< _CharT, _Traits >& ss, const fh_node& n )
//         {
//             string t;
//             n->encode( t );
//             ss << t;
//             return ss;
//         }
//         template < class _CharT, class _Traits >
//         std::basic_ostream< _CharT, _Traits >&
//         operator>>( std::basic_ostream< _CharT, _Traits >& ss, const fh_node& n )
//         {
//             string t;
//             n->decode( t );
//             ss << t;
//             return ss;
//         }
        
        /******************************/
        /******************************/
        /******************************/

        FERRISEXP_API bool operator==(const Statement& __x, const Statement& __y);
        FERRISEXP_API bool operator!=(const Statement& __x, const Statement& __y);
        class FERRISEXP_API Statement
            :
            public Handlable
        {
            friend class       StatementIterator;
            librdf_statement*  m_statement;
            bool               m_ownStatement;
            
            Statement( librdf_statement* m_statement, bool takeOwnership = true );
            Statement( const Statement& );
            Statement& operator=( const Statement& );

            friend bool operator==(const Statement& __x, const Statement& __y);
            
        public:

            Statement( );
            Statement( fh_node sub, fh_node pred, fh_node obj );
            
            virtual ~Statement();
            librdf_statement* getRAW() const;

            void setSubject( fh_node n );
            void setPredicate( fh_node n );
            void setObject( fh_node n );

            fh_node getSubject();
            fh_node getPredicate();
            fh_node getObject();

            void clear();
            bool isComplete();
            std::string toString();
            bool match( fh_statement partial_st );

            std::string toNTripleString( fh_model m );
            std::string toRDFXMLString( fh_model m );
            std::string toXString( fh_model m, const std::string& format );
        };

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        FERRISEXP_API bool operator==(const URI& __x, const URI& __y);
        FERRISEXP_API bool operator!=(const URI& __x, const URI& __y);
        class FERRISEXP_API URI
            :
            public Handlable
        {
            friend class ImplWorld;
            NOT_COPYABLE( URI );

            librdf_uri* m_uri;

            URI( librdf_uri* u );

            friend bool operator==(const URI& __x, const URI& __y);
            
        public:

            explicit URI( const std::string& uri );
            virtual ~URI();
            librdf_uri* getRAW() const;

            std::string toString();
            std::string getPath();
            bool        isFile();
        };

        struct fh_node_compare
        {
            bool operator()( fh_node n1, fh_node n2 ) const
                {
                    return n1->toString() < n2->toString();
                }
        };
        
        namespace Wordnet
        {
            FERRISEXP_API extern std::string RDFSCHEMA;
            FERRISEXP_API extern std::string WNPFX;

            typedef std::list< fh_node > nodelist_t;

            FERRISEXP_API fh_node& hasWord();
            FERRISEXP_API fh_node& hasSynSet();
            FERRISEXP_API fh_node& NounSynSet();
            FERRISEXP_API fh_node& Type();
            FERRISEXP_API fh_node& hasGloss();
            FERRISEXP_API fh_node& hyponymOf();
            FERRISEXP_API fh_node& partMeronymOf();
            
            FERRISEXP_API nodelist_t& getSynSet( fh_model m, nodelist_t& ret, const std::string& word );
            FERRISEXP_API std::string getNodeID( fh_node n );
            FERRISEXP_API std::string getWordNodeAsString( fh_node n );
        };
        
        
    };

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    namespace redlandea
    {
        FERRISEXP_API std::string getAttrPrefix();
        FERRISEXP_API std::string getPredicateURI( const std::string& rdn );
        FERRISEXP_API std::string stripPredicateURIPrefix( const std::string& pred );

        /**
         * After a file is moved/copied outside of libferris this call can attach
         * EA which is stored in myrdf.db to the new version of the context.
         * eg. mv foo foo1; myrdfSmush( Resolve( foo1 ), foo );
         */
        FERRISEXP_API void myrdfSmush( fh_context existingc, const std::string& oldpath, bool unifyUUIDNodes = false );
        FERRISEXP_API void myrdfSmush( fh_context existingc, fh_context oldc, bool unifyUUIDNodes = false );

        /**
         * Merge the myrdf metadata for the given list of URLs making each URL
         * see the union of this metadata. If an EA is bound for two URLs
         * each URL will see its old value. So this really only imports the
         * EAs which are not bound for each respective URL in the set from all
         * the others in the set.
         */
        FERRISEXP_API void myrdfSmushUnion( const stringlist_t& srcs );

        /**
         * Filesystem objects have a UUID node which has metadata attached to it.
         * Conceptually it is something like this
         *
         * file-url uuid-pred uuidnode
         * uuidnode redland-ea-pred RedlandEASubjectNodeX
         * RedlandEASubjectNodeX eaname eavalue
         *
         * Get the uuidnode with ensureUUIDNode().
         * Get the RedlandEASubjectNodeX to get all the out-of-band EA stored in myrdf
         * using ensureRedlandEASubjectNode()
         * 
         * The uuid-pred link can be made implicitly sometimes by libferris so its
         * best to call tryToGetUUIDNode( fh_context ) to find the UUID node
         * associated with a context if one exists.
         */
        FERRISEXP_API ::Ferris::RDF::fh_node tryToGetUUIDNode( fh_context c );

        /**
         * Like tryToGetUUIDNode() but if no such link exists it is created
         * automatically.
         */
        FERRISEXP_API ::Ferris::RDF::fh_node ensureUUIDNode( fh_context c );
        FERRISEXP_API ::Ferris::RDF::fh_node ensureUUIDNode( ::Ferris::RDF::fh_model m, fh_context c );

        /**
         * Get the last time_t any of the RDF data relating to this uuidnode was
         * changed.
         */
        FERRISEXP_API time_t getUUIDNodeModificationTime( ::Ferris::RDF::fh_node n );

        /**
         * Record the time_t that an update for any RDF data relatiing to this
         * uuidnode has changed.
         */
        FERRISEXP_API void setUUIDNodeModificationTime( ::Ferris::RDF::fh_node n,
                                                        time_t t = 0 );
        FERRISEXP_API void setRedlandEASubjectNode( ::Ferris::RDF::fh_node n,
                                                    time_t t = 0 );
        
        /**
         * Try to get the subject node for a context where EAName EAValue predicate object
         * pairings are stored.
         *
         * @see ensureRedlandEASubjectNode()
         */
        FERRISEXP_API ::Ferris::RDF::fh_node tryToGetRedlandEASubjectNode( fh_context c );
        FERRISEXP_API ::Ferris::RDF::fh_node ensureRedlandEASubjectNode( fh_context c );
        
        /**
         * This is mainly for applications like format converters which need to
         * setup uuid-pred links themselves.
         * Applications should use tryToGetUUIDNode() and ensureUUIDNode().
         */
        FERRISEXP_API ::Ferris::RDF::fh_node& uuidPredNode();
        /**
         * Used for creating a quick
         * smushsetid-filename smushCacheUUIDPredNode() uuidnode
         * cache for fast smushing on filename.
         */
        FERRISEXP_API ::Ferris::RDF::fh_node& smushCacheUUIDPredNode();
        

        /**
         * This is mainly for applications like format converters which need to
         * setup uuid-pred links themselves.
         * Applications should use tryToGetUUIDNode() and ensureUUIDNode().
         */
        FERRISEXP_API ::Ferris::RDF::fh_node& uuidOutOfBandPredNode();
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        
        class FERRISEXP_API SmushSet
            :
            public Handlable
        {
        public:
            virtual ~SmushSet();

            const std::string& getName() const;
            void setName( const std::string& n );

            typedef std::map< std::string, fh_regex > m_smushes_t;
            const m_smushes_t& getSmushes() const;
            const std::string& getRegexString();
            fh_rex getRegex();
            void setSmushes( m_smushes_t& v );
            
        private:
            std::string m_name;
            m_smushes_t m_smushes;
            std::string m_regexString;
            fh_rex      m_rex;
            friend class TreeSmushing;
            fh_TreeSmushing m_treeSmushing;
            SmushSet( fh_TreeSmushing p, const std::string& name );
            
            friend class boost::serialization::access;
        };

        FERRISEXP_API fh_TreeSmushing getDefaultImplicitTreeSmushing();
        
        class FERRISEXP_API TreeSmushing
            :
            public Handlable
        {
        public:
            TreeSmushing();
            virtual ~TreeSmushing();

            typedef std::map< std::string, fh_SmushSet > m_smushSets_t;

            const m_smushSets_t& getAll() const;
            fh_SmushSet getSmush( const std::string& n ) const;
            void swap( fh_TreeSmushing other );
            fh_SmushSet newSmush( const std::string& name );
            void sync();

            template< class ArchiveClassT >
            void sync( ArchiveClassT& archive );

            
        private:
            bool m_shouldSave;
            m_smushSets_t m_smushSets;

            TreeSmushing( bool shouldSave );
            friend fh_TreeSmushing getDefaultImplicitTreeSmushing();

            static std::string getBoostSerializePath();
            
            friend class SmushSet;
            void updateSmushName( const std::string& oldname, const std::string& newname );

            friend class boost::serialization::access;
            template<class Archive>
            void serialize(Archive & ar, const unsigned int version)
                {
                    ar & m_smushSets;
                }
        };
        

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        
    };
    
};
#endif
