/******************************************************************************
*******************************************************************************
*******************************************************************************

    ferris internal metadata broker
    Copyright (C) 2007 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: ferris-internal-metadata-broker.cpp,v 1.4 2007/08/31 21:30:13 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <Ferris/Ferris.hh>
#include <Ferris/Ferris_private.hh>
#include <Ferris/MatchedEAGenerators.hh>
#include <Ferris/FerrisEAGeneratorPlugin_private.hh>
#include <Ferris/Image_private.hh>
#include <Ferris/MetadataServer_private.hh>

#include <popt.h>
#include <signal.h>

#include <dbus-c++/dbus.h>
#include <dbus-c++/glib-integration.h>
#include <dbus-c++/pendingcall.h>
#include <dbus-c++/types.h>
#include <dbus-c++/message.h>
#include "DBusGlue/ferris-internal-metadata-broker-server-glue.h"
#include "DBusGlue/ferris-internal-metadata-worker-client-glue.h"

using namespace std;
using namespace Ferris;

const string PROGRAM_NAME = "ferris-internal-metadata-broker";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

#define USE_PENDINGCALL


#ifndef USE_PENDINGCALL

#include <boost/thread.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
using namespace boost;
using namespace boost::detail::thread;

#endif


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


static const char* DBUS_SERVER_NAME = "net.sf.witme.Libferris.Metadata.Broker";
static const char* DBUS_SERVER_PATH = "/net/sf/witme/libferris/Metadata";
static const char* WORKER_OBJECT = "net.sf.witme.Libferris.Metadata.Worker";
static string WORKER_PREFIX = (string)WORKER_OBJECT + ".";

int RETURN_CODE = 0;
bool running = true;

namespace Ferris
{
    guint32 glib_idle_timeout_Interval = 30 * 60;
    void resetup_glib_idle_to_close_callback_function()
    {
        resetup_glib_idle_to_close_callback_function( glib_idle_timeout_Interval, &running );
    }
};


namespace Ferris
{
    class MetadataWorker
        :
        public net::sf::witme::Libferris::Metadata::Worker,
        public DBus::IntrospectableProxy,
        public DBus::ObjectProxy
    {
    public:

        MetadataWorker( DBus::Connection& connection, const char* path, const char* name )
            :
            DBus::ObjectProxy(connection, path, name)
            {}
    };
};


namespace Ferris
{
    typedef ::DBus::Slot<void, ::DBus::PendingCall&> AsyncCallSlot_t;
};



namespace Ferris
{
    class MetadataBrokerDispatchInformation
        :
        public Handlable
    {
        typedef map< string, stringlist_t > m_plugins_t;

        string getEAGeneratorName( const string& earl,
                                   const string& rdn,
                                   bool writable )
            {
                LG_MDSERV_D << "getEAGeneratorName(called) earl:" << earl
                            << " rdn:" << rdn
                            << " writable:" << writable
                            << endl;
                string ret = "";

                try
                {
                    fh_context ctx = Resolve( earl );
                    LG_MDSERV_D << "resolve()ed ctx:" << ctx->getURL() << endl;
                    
                    StaticEAGenFactorys_t fac;
                    getOutOfProcess_EAGeneratorsStaticFactories( fac );

                    LG_MDSERV_D << "Looking at factories - start... fac.sz:" << fac.size() << endl;
                    {
                        StaticEAGenFactorys_t::iterator end = fac.end();
                        for( StaticEAGenFactorys_t::iterator fi = fac.begin(); fi != end; ++fi )
                        {
                            Context::fh_MatchedEAGeneratorFactory f = *fi;
//                             if( f->hasState() )
//                             {
//                                 f = f->clone();
//                             }

                            if( f->hasInterest( ctx ) )
                            {
                                if( StaticGModuleMatchedEAGeneratorFactory* fd =
                                    dynamic_cast<StaticGModuleMatchedEAGeneratorFactory*>(GetImpl( f )))
                                {
                                    string sn = fd->getShortName();
                                    LG_MDSERV_D << "getEAGeneratorName(called) earl:" << earl
                                                << " rdn:" << rdn
                                                << " writable:" << writable
                                                << " sn:" << sn
                                                << endl;
                                    return sn;
                                }
                            }
                        }
                    }
                }
                catch( exception& e )
                {
                    LG_MDSERV_I << "getEAGeneratorName(called) earl:" << earl
                                << " rdn:" << rdn
                                << " writable:" << writable
                                << " e:" << e.what()
                                << endl;
                }
                
                return ret;
            }
        
        string getImageEAGeneratorName( const string& earl,
                                        const string& rdn,
                                        bool writable )
            {
                LG_MDSERV_D << "getImageEAGeneratorName(called) earl:" << earl
                            << " rdn:" << rdn
                            << " writable:" << writable
                            << endl;

                if( isImageMetadataAttribute( rdn ) )
                {
                    LG_MDSERV_D << "getImageEAGeneratorName(not image metadata) earl:" << earl
                                << " rdn:" << rdn
                                << " writable:" << writable
                                << endl;
                    return "";
                }
                
                
                string ret = "";
                string nameExtension = "";
                string::size_type pos = earl.rfind('.');
                if( string::npos != pos )
                {
                    nameExtension = earl.substr( pos+1 );
                }
                nameExtension = tolowerstr()( nameExtension );

                LG_MDSERV_D << "getImageEAGeneratorName() earl:" << earl
                            << " nameExtension:" << nameExtension
                            << " writable:" << writable
                            << endl;
                
                if( !nameExtension.empty() )
                {
                    s_ImageEAGeneratorsExtensionToShortName_t& genmap = getOutOfProcess_ImageEAGeneratorsExtensionToShortName();
                    s_ImageEAGeneratorsExtensionToShortName_t::iterator giter = genmap.find( nameExtension );

                    LG_MDSERV_D << "getImageEAGeneratorName() earl:" << earl
                                << " nameExtension:" << nameExtension
                                << " writable:" << writable
                                << " genmap.sz:" << genmap.size()
                                << " found:" << ( giter != genmap.end() )
                                << endl;
                    
                    if( giter != genmap.end() )
                    {
                        string pshortname = giter->second.first;
                        bool   pwritable  = giter->second.second;

                        if( !writable || (writable && pwritable) )
                            ret = pshortname;
                    }
                }
                
                return ret;
            }
        
        
    public:
        MetadataBrokerDispatchInformation()
            {
            }
        ~MetadataBrokerDispatchInformation()
            {
            }
        
        string getServiceName( const ::DBus::String& earl, const ::DBus::String& name, bool writable = false )
            {
                LG_MDSERV_D << "getServiceName() earl:" << earl << " name:" << name << " writable:" << writable << endl;
                
                string ret = "";
                ret = getImageEAGeneratorName( earl, name, writable );
                if( !ret.empty() )
                {
                    LG_MDSERV_D << "getServiceName(img) earl:" << earl
                                << " name:" << name
                                << " writable:" << writable
                                << " ret:" << ret
                                << endl;
                    return WORKER_PREFIX + ret;
                }


                ret = getEAGeneratorName( earl, name, writable );
                if( !ret.empty() )
                {
                    LG_MDSERV_D << "getServiceName(ea) earl:" << earl
                                << " name:" << name
                                << " writable:" << writable
                                << " ret:" << ret
                                << endl;
                    return WORKER_PREFIX + ret;
                }
                

                LG_MDSERV_D << "getServiceName(done) earl:" << earl
                            << " name:" << name
                            << " writable:" << writable
                            << " ret:" << ret
                            << endl;
                return ret;
            }
    };
    FERRIS_SMARTPTR( MetadataBrokerDispatchInformation, fh_MetadataBrokerDispatchInformation );
};


namespace Ferris
{
    using namespace DBus;
    
    typedef DBus::Int32 reqid_t;


#ifdef USE_PENDINGCALL
    
    class MetadataBroker
        :
        public net::sf::witme::Libferris::Metadata::Broker,
        public DBus::IntrospectableAdaptor,
        public DBus::ObjectAdaptor
    {
        reqid_t m_nextReqID;
        fh_MetadataBrokerDispatchInformation m_mdDispatchInfo;
        
    public:
        typedef std::list< DBus::PendingCall > m_pendingGetCalls_t;
        m_pendingGetCalls_t m_pendingGetCalls;

        typedef std::list< DBus::PendingCall > m_pendingPutCalls_t;
        m_pendingPutCalls_t m_pendingPutCalls;
        
        void asyncGetFireReply( ::DBus::PendingCall& pc,
                                reqid_t reqid,
                                string earl,
                                string attr )
            {
                LG_MDSERV_D << "asyncGetFireReply(called)"
                            << " reqid:" << reqid
                            << " attr:" << attr
                            << " earl:" << earl
                            << endl;

                ::DBus::Message m = pc.steal_reply();
                LG_MDSERV_D << "reqid:" << reqid << " is_error():" << m.is_error() << endl;

                if( m.is_error() )
                {
                    const ::DBus::ErrorMessage em = reinterpret_cast<const ::DBus::ErrorMessage&>(m);
                    LG_MDSERV_D << "em.name:" << em.name() << endl;

                    
//                     // The other end died or didn't return anything...
//                     if( !strcmp( em.name(), "org.freedesktop.DBus.Error.NoReply" ))
//                     {
//                         cerr << "The other end died or didn't return anything..." << endl;

//                         cerr << "Trying to ping service..." << endl;

//                         static stringset_t earlGuard;

//                         if( earlGuard.end() == earlGuard.find( earl ) )
//                         {
//                             earlGuard.insert( earl );
//                             string workerName = m_mdDispatchInfo->getServiceName( earl, attr, false );
//                             int timeout = 200;
//                             cerr << "attr:" << attr << " "
//                                  << "earl:" << earl << endl
//                                  << "...Trying to send a ping to workerName:" << workerName
//                                  << endl;
//                             try
//                             {
// //                                Ferris::MetadataWorker client( m_conn, DBUS_SERVER_PATH, workerName.c_str() );
// //                                cerr << "ping:" << client.ping() << endl;

//                                 DBus::CallMessage call( workerName.c_str(), DBUS_SERVER_PATH, WORKER_OBJECT, "ping" );
//                                 Message m = m_conn.send_blocking(call, 1000);
//                                 ::DBus::MessageIter ri = m.reader();
//                                 ::DBus::Int32 argout;
//                                 ri >> argout;
//                                 cerr << "argout:" << argout << endl;
                                
//                                 asyncGet( earl, attr );
//                                 return;
//                             }
//                             catch( exception& e )
//                             {
//                                 cerr << "ping test error:" << e.what() << endl;
//                             }
//                             earlGuard.erase( earl );
//                         }
//                     }
                    int ecode = 1;
                    string ename = em.name();
                    string edesc = "";
                    DBus::MessageIter ri = m.reader();
                    ri >> edesc;

                    LG_MDSERV_I << "asyncGetFireReply(error)"
                                << " reqid:" << reqid
                                << " attr:" << attr
                                << " earl:" << earl
                                << " ecode:" << ecode
                                << " reason:" << edesc
                                << endl;
                    this->asyncGetFailed( reqid, earl, ecode, ename, edesc );
                }
                else
                {
                    ::DBus::MessageIter ri = m.reader();
                    std::vector< DBus::Byte > value;
                    ri >> value;

                    LG_MDSERV_D << "asyncGetFireReply(ok)"
                                << " reqid:" << reqid
                                << " attr:" << attr
                                << " earl:" << earl
                                << " value:" << value
                                << endl;

                    this->asyncGetResult( reqid, earl, attr, value );
                }
            }

        void asyncPutFireReply( ::DBus::PendingCall& pc,
                                reqid_t reqid,
                                string earl,
                                string attr )
            {
                LG_MDSERV_D << "asyncPutFireReply(called)"
                            << " reqid:" << reqid
                            << " attr:" << attr
                            << " earl:" << earl
                            << endl;
                ::DBus::Message m = pc.steal_reply();
                
                if( m.is_error() )
                {
                    const ::DBus::ErrorMessage em = reinterpret_cast<const ::DBus::ErrorMessage&>(m);
                    cerr << "em.name:" << em.name() << endl;

                    // The other end died or didn't return anything...
                    if( !strcmp( em.name(), "org.freedesktop.DBus.Error.NoReply" ))
                    {
                        LG_MDSERV_D << "asyncPutFireReply(called)"
                                    << "The other end died or didn't return anything..." << endl
                                    << " reqid:" << reqid
                                    << " attr:" << attr
                                    << " earl:" << earl
                                    << endl;
                    }

                    int ecode = 1;
                    string ename = em.name();
                    string edesc = "";
                    DBus::MessageIter ri = m.reader();
                    ri >> edesc;
                    LG_MDSERV_I << "asyncPutFireReply(error)"
                                << " reqid:" << reqid
                                << " attr:" << attr
                                << " earl:" << earl
                                << " ecode:" << ecode
                                << " reason:" << edesc
                                << endl;
                    
                    this->asyncPutFailed( reqid, earl, ecode, ename, edesc );
                }
                else
                {
                    LG_MDSERV_D << "asyncPutFireReply(ok)"
                                << " reqid:" << reqid
                                << " attr:" << attr
                                << " earl:" << earl
                                << endl;
                    this->asyncPutCommitted( reqid, earl, attr );
                }
            }
        
        
    private:
        
        DBus::Connection m_conn;
        
        DBus::Connection& getConnection()
            {
                return m_conn;
            }

        
    public:
    
        MetadataBroker( DBus::Connection& connection );
        ~MetadataBroker();

        virtual ::DBus::Int32 Random(  )
            {
            }
        
//         DBus::String get( const DBus::String& earl, const DBus::String& name );
//         void set( const DBus::String& earl, const DBus::String& name, const DBus::String& value );

        DBus::Int32 asyncGet( const ::DBus::String& earl, const ::DBus::String& name );
        DBus::Int32 asyncPut( const ::DBus::String& earl,
                              const ::DBus::String& name,
                              const std::vector< ::DBus::Byte >& value );
        
    };
    
    MetadataBroker::MetadataBroker( DBus::Connection& connection )
        :
        DBus::ObjectAdaptor( connection, DBUS_SERVER_PATH ),
        m_conn( connection ),
        m_nextReqID( 0 ),
        m_mdDispatchInfo( new MetadataBrokerDispatchInformation() )
    {
        LG_MDSERV_D << "MetadataBroker()" << endl;
    }

    MetadataBroker::~MetadataBroker()
    {
        LG_MDSERV_D << "~MetadataBroker()" << endl;
    }

    

    
    struct asyncXCallback
        :
        public ::DBus::Callback_Base<void,::DBus::PendingCall&>
    {
    public:

        MetadataBroker* m_broker;
        typedef long reqid_t;
        reqid_t m_reqid;
        string m_earl;
        string m_attr;
    
        asyncXCallback( MetadataBroker* m_broker,
                        reqid_t m_reqid,
                        string m_earl,
                        string m_attr )
            :
            m_broker( m_broker ),
            m_reqid( m_reqid ),
            m_earl( m_earl ),
            m_attr( m_attr )
            {
            }
    };

    struct asyncGetCallback
        :
        public asyncXCallback
    {
        asyncGetCallback( MetadataBroker* m_broker,
                          reqid_t m_reqid,
                          string m_earl,
                          string m_attr )
            :
            asyncXCallback( m_broker, m_reqid, m_earl, m_attr )
            {}

        void call( ::DBus::PendingCall& param ) const
            {
                return m_broker->asyncGetFireReply( param, m_reqid, m_earl, m_attr );
            }
    };

    struct asyncPutCallback
        :
        public asyncXCallback
    {
        asyncPutCallback( MetadataBroker* m_broker,
                          reqid_t m_reqid,
                          string m_earl,
                          string m_attr )
            :
            asyncXCallback( m_broker, m_reqid, m_earl, m_attr )
            {}

        void call( ::DBus::PendingCall& param ) const
            {
                return m_broker->asyncPutFireReply( param, m_reqid, m_earl, m_attr );
            }
    };
    
    
    
    DBus::Int32
    MetadataBroker::asyncGet( const ::DBus::String& earl, const ::DBus::String& name )
    {
        resetup_glib_idle_to_close_callback_function();
        LG_MDSERV_D << "MetadataBroker::asyncGet(called) earl:" << earl << endl;
        
        DBus::Int32 ret = ++m_nextReqID;
        
        string workerName = m_mdDispatchInfo->getServiceName( earl, name, false );
        LG_MDSERV_D << "MetadataBroker::asyncGet() earl:" << earl
                    << " rdn:" << name << " workerName:" << workerName
                    << endl;
        try
        {
            LG_MDSERV_D << "MetadataBroker::asyncGet() earl:" << earl
                        << " sending ping to workerName:" << workerName << endl;
            DBus::CallMessage call( workerName.c_str(), DBUS_SERVER_PATH, WORKER_OBJECT, "ping" );
            Message m = m_conn.send_blocking(call, 1000);
            ::DBus::MessageIter ri = m.reader();
            ::DBus::Int32 argout;
            ri >> argout;
            LG_MDSERV_D << "MetadataBroker::asyncGet() earl:" << earl
                        << " got ping reply:" << argout << " from workerName:" << workerName << endl;
        }
        catch( exception& e )
        {
            LG_MDSERV_D << "MetadataBroker::asyncGet() earl:" << earl
                        << " error sending ping to workerName:" << workerName
                        << " reason:" << e.what()
                        << endl;
        }
        
        
        int timeout = 10000;
        DBus::CallMessage call( workerName.c_str(), DBUS_SERVER_PATH, WORKER_OBJECT, "get" );
        ::DBus::MessageIter wi = call.writer();
        wi << earl;
        wi << name;
        LG_MDSERV_D << "MetadataBroker::asyncGet(calling async_send) earl:" << earl
                    << " rdn:" << name << " workerName:" << workerName
                    << endl;
        DBus::PendingCall pc = m_conn.send_async( call, timeout );
//        static DBus::PendingCall pc2 = pc;
        
        cerr << "setting asyncGetFireReply slot..." << endl;
//        pc.slot() = new DBus::Callback<MetadataBroker, void, DBus::PendingCall&>( this, &MetadataBroker::asyncGetFireReply );
        pc.slot() = new asyncGetCallback( this, ret, earl, name );

        
        m_pendingGetCalls.push_back( pc );
//        m_pendingGetCalls.push_back( pc );
        
//        cerr << "calling block()..." << endl;
//        pc.block();

        LG_MDSERV_D << "MetadataBroker::asyncGet(done) earl:" << earl
                    << " rdn:" << name << " workerName:" << workerName
                    << endl;
        
        return ret;
    }
    
    DBus::Int32
    MetadataBroker::asyncPut( const ::DBus::String& earl,
                              const ::DBus::String& name,
                              const std::vector< ::DBus::Byte >& value )
    {
        resetup_glib_idle_to_close_callback_function();
        LG_MDSERV_D << "MetadataBroker::asyncPut(called) earl:" << earl << endl;

        DBus::Int32 ret = ++m_nextReqID;

        string workerName = m_mdDispatchInfo->getServiceName( earl, name, true );
        int timeout = 5000;
        DBus::CallMessage call( workerName.c_str(), DBUS_SERVER_PATH, WORKER_OBJECT, "put" );
        
        ::DBus::MessageIter wi = call.writer();
        wi << earl;
        wi << name;
        wi << value;
        LG_MDSERV_D << "MetadataBroker::asyncPut(calling send_async) earl:" << earl
                    << " rdn:" << name
//                    << " value:" << value
                    << " workerName:" << workerName
                    << endl;
        DBus::PendingCall pc = m_conn.send_async( call, timeout );

        pc.slot() = new asyncPutCallback( this, ret, earl, name );
        m_pendingPutCalls.push_back( pc );
        
        LG_MDSERV_D << "MetadataBroker::asyncPut(returning) earl:" << earl << endl;
        return ret;
    }

#else

    struct JobData
        :
        public Handlable
    {
        DBus::Int32 reqid;
        DBus::String earl;
        DBus::String name;
        std::vector< ::DBus::Byte > value;
        
        JobData( DBus::Int32 reqid,
                 DBus::String earl,
                 DBus::String name,
                 const std::vector< ::DBus::Byte > value )
            :
            reqid( reqid ),
            earl( earl ),
            name( name ),
            value( value )
            {
            }
        JobData( DBus::Int32 reqid,
                 DBus::String earl,
                 DBus::String name )
            :
            reqid( reqid ),
            earl( earl ),
            name( name )
            {
            }
    };
    FERRIS_SMARTPTR( JobData, fh_JobData );
    
    
    class MetadataBroker
        :
        public net::sf::witme::Libferris::Metadata::Broker,
        public DBus::IntrospectableAdaptor,
        public DBus::ObjectAdaptor
    {
        reqid_t m_nextReqID;
        fh_MetadataBrokerDispatchInformation m_mdDispatchInfo;
        
    public:

        enum
        {
            THREAD_COUNT = 10
        };
        thread_group m_threads;

        
        typedef list< fh_JobData > m_jobQ_t;
        mutex m_getQMutex;
        mutex m_putQMutex;
        m_jobQ_t m_getQ;
        m_jobQ_t m_PutQ;
        
    private:
        
        DBus::Connection m_conn;
        
        DBus::Connection& getConnection()
            {
                return m_conn;
            }

    public:
    
        MetadataBroker( DBus::Connection& connection );
        ~MetadataBroker();

        virtual ::DBus::Int32 Random(  )
            {
            }
        
//         DBus::String get( const DBus::String& earl, const DBus::String& name );
//         void set( const DBus::String& earl, const DBus::String& name, const DBus::String& value );

        DBus::Int32 asyncGet( const ::DBus::String& earl, const ::DBus::String& name );
        DBus::Int32 asyncPut( const ::DBus::String& earl,
                              const ::DBus::String& name,
                              const std::vector< ::DBus::Byte >& value );

        void thread_get_function();
    };
    
    MetadataBroker::MetadataBroker( DBus::Connection& connection )
        :
        DBus::ObjectAdaptor( connection, DBUS_SERVER_PATH ),
        m_conn( connection ),
        m_nextReqID( 0 ),
        m_mdDispatchInfo( new MetadataBrokerDispatchInformation() )
    {
        
        LG_MDSERV_D << "MetadataBroker()" << endl;
        for( int i=0; i < THREAD_COUNT; ++i )
        {
//             boost::function<void> f;
//             f = &MetadataBroker::thread_get_function;
            m_threads.create_thread( boost::bind(&MetadataBroker::thread_get_function, this) );
        }
        
    }

    MetadataBroker::~MetadataBroker()
    {
        LG_MDSERV_D << "~MetadataBroker()" << endl;
    }

    void
    MetadataBroker::thread_get_function()
    {
        while( true )
        {
            fh_JobData jd = 0;
            
            {
                scoped_lock<mutex> lk( m_getQMutex );
                if( !m_getQ.empty() )
                {
                    jd = m_getQ.front();
                    m_getQ.pop_front();
                }
            }

            if( !jd )
            {
                sleep( 1 );
                continue;
            }

            DBus::Int32 reqid = jd->reqid;
            string earl = jd->earl;
            string rdn = jd->name;
            try
            {
                string workerName = m_mdDispatchInfo->getServiceName( earl, rdn, false );
                LG_MDSERV_D << "MetadataBroker::thread() earl:" << earl
                            << " rdn:" << rdn << " workerName:" << workerName
                            << endl;

                
                Ferris::MetadataWorker client( m_conn, DBUS_SERVER_PATH, workerName.c_str() );
                LG_MDSERV_D << "MetadataBroker::thread(have client) earl:" << earl
                            << " rdn:" << rdn << " workerName:" << workerName
                            << endl;
                
                const std::vector< ::DBus::Byte >& value = client.get( earl, rdn );
                LG_MDSERV_D << "MetadataBroker::thread() earl:" << earl
                            << " rdn:" << rdn << " workerName:" << workerName
                            << endl;
                LG_MDSERV_D << "asyncGetFireReply(ok)"
                            << " reqid:" << reqid
                            << " rdn:" << rdn
                            << " earl:" << earl
                            << " value:" << value
                            << endl;
                this->asyncGetResult( reqid, earl, rdn, value );
            }
            catch( exception& e )
            {
                int ecode = 1;
                string ename = "";
                string edesc = e.what();

                LG_MDSERV_I << "asyncGetFireReply(error)"
                            << " reqid:" << reqid
                            << " rdn:" << rdn
                            << " earl:" << earl
                            << " ecode:" << ecode
                            << " reason:" << edesc
                            << endl;
                this->asyncGetFailed( reqid, earl, ecode, ename, edesc );
            }
        }
    }
    

    
    DBus::Int32
    MetadataBroker::asyncGet( const ::DBus::String& earl, const ::DBus::String& name )
    {
        resetup_glib_idle_to_close_callback_function();
        LG_MDSERV_D << "MetadataBroker::asyncGet(called) earl:" << earl << endl;
        
        DBus::Int32 reqid = ++m_nextReqID;

        fh_JobData jd = new JobData( reqid, earl, name );
        {
            scoped_lock<mutex> lk( m_getQMutex );
            m_getQ.push_back( jd );
        }
        
        LG_MDSERV_D << "MetadataBroker::asyncGet(done) earl:" << earl
                    << " rdn:" << name << endl;

        return reqid;
    }
    
    DBus::Int32
    MetadataBroker::asyncPut( const ::DBus::String& earl,
                              const ::DBus::String& name,
                              const std::vector< ::DBus::Byte >& value )
    {
    }


#endif    
};



////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

using namespace Ferris;
DBus::Glib::BusDispatcher dispatcher;


void niam( int sig )
{
    running = false;
//	dispatcher.leave();
}

int main( int argc, const char** argv )
{
	signal(SIGTERM, niam);
	signal(SIGINT, niam);

	DBus::default_dispatcher = &dispatcher;
    dispatcher.attach(NULL);

    setForceOutOfProcessMetadataOff( true );
    
    unsigned long ShowVersion        = 0;
    const char*   PluginNames_CSTR   = 0;
    const char*   FormatDisplayCSTR  = 0;

    struct poptOption optionsTable[] = {

        { "version", 0, POPT_ARG_NONE, &ShowVersion, 0,
          "show version information and quit", 0 },

//         { "plugin-names", 0, POPT_ARG_STRING, &PluginNames_CSTR, 0,
//           "comma seperated list of plugins to use for metadata handling", 0 },

        /*
         * Standard Ferris options
         */
//        FERRIS_POPT_OPTIONS
        POPT_AUTOHELP
        POPT_TABLEEND
    };
    poptContext optCon;


    optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* ...");

    /* Now do options processing, get portname */
    char c=-1;
    while ((c = poptGetNextOpt(optCon)) >= 0)
    {}

    if (argc < 1 )
    {
        cerr << "Arguments incorrect!" << endl;
//        poptPrintHelp(optCon, stderr, 0);
        poptPrintUsage(optCon, stderr, 0);
        exit(1);
    }
    
    if( ShowVersion )
    {
        cout << "ferris-internal-metadata-worker" << nl
             << "version: $Id: ferris-internal-metadata-broker.cpp,v 1.4 2007/08/31 21:30:13 ben Exp $\n"
             << "Written by Ben Martin, aka monkeyiq" << nl
             << nl
             << "Copyright (C) 2007 Ben Martin" << nl
             << "This is free software; see the source for copying conditions.  There is NO\n"
             << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."
             << endl;
        exit(0);
    }

    
    try
    {
        resetup_glib_idle_to_close_callback_function();

        LG_MDSERV_D << "broker starting up..." << endl;
        
        DBus::Connection conn = DBus::Connection::SessionBus();
        conn.request_name( DBUS_SERVER_NAME );
        MetadataBroker server( conn );

        while( running )
            g_main_iteration( true );

    }
    catch( NoSuchContextClass& e )
    {
        LG_MDSERV_W << "broker error:" << e.what() << endl;
        cerr << "e:" << e.what() << endl;
        exit(1);
    }
    catch( exception& e )
    {
        LG_MDSERV_W << "broker error:" << e.what() << endl;
        cerr << "cought:" << e.what() << endl;
        exit(1);
    }
        
    LG_MDSERV_D << "broker exiting..." << endl;
    poptFreeContext(optCon);
    cout << flush;
    return RETURN_CODE;
}
