/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: libferriscurl.cpp,v 1.6 2006/08/13 11:41:36 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/
//
//
// Note that RFC 0959 is the FTP one.
//
//
#include <config.h>

#include <Ferris_private.hh>
#include <FerrisContextPlugin.hh>
#include <Trimming.hh>
#include <General.hh>

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#include <SmartPtr.h>
#include <Singleton.h>
#include <Factory.h>
#include <Functor.h>

#include <string>
#include <map>
#include <vector>

#include <sys/utsname.h>
#include <time.h>

#include <sigc++/sigc++.h>
#include <sigc++/slot.h>
#include <sigc++/bind.h>
#include <sigc++/object.h>
#include <sigc++/object_slot.h>

#define CURL_LEAVES_SIG_ALARM_ON
#ifdef CURL_LEAVES_SIG_ALARM_ON
#include <signal.h>
#endif

using namespace std;




namespace Ferris
{
    /*
     * Just contains the server name and path of the object we want
     */
    struct FtpDetails;

    /*
     * Context subclass for ftp/http contexts
     */
    class curlFtpContext;

    /*
     * Stores the HTTP headers or FTP commands and manages network IO
     */
    class FerrisCurlState;

    /*
     * FIXME: why?
     */
    class FerrisCurlClass;
    

    
    extern "C"
    {
        FERRISEXP_EXPORT fh_context Brew( RootContextFactory* rf )
            throw( RootContextCreationFailed );
    };

    struct FtpDetails
    {
        string serv;
        string path;

        FtpDetails( const string& _serv = "", const string& _path = "" )
            :
            serv( _serv ),
            path( _path )
            {}

        void reset()
            {
                serv = "";
                path = "";
            }
    };


    
    /*
     * /ftp/localhost/pub/somewhere/anotherdir
     *  |       |      |-------+------|
     *  x     serv     path
     */
    const FtpDetails& getFtpDetails( fh_context c )
    {
        static FtpDetails d;
        typedef list< fh_context > flattree_t;
        flattree_t ftree;

        d.reset();

        fh_context pc = c;
        while( pc->isParentBound() )
        {
            ftree.push_front( pc );
            pc = pc->getParent();
        }

        flattree_t::iterator iter = ftree.begin();
        
        if( iter != ftree.end() )
        {
            d.serv = (*iter)->getDirName();

            d.path = "/";
            while( ++iter != ftree.end() )
            {
                d.path += "/";
                d.path += (*iter)->getDirName();
            }
        }
        else
        {
            fh_stringstream ss;
            ss << "Can not determine the server name for context:"
               << c->getDirPath();
            Throw_FerrisCurlServerNameNotFound( tostr(ss), 0 );
        }
        
        LG_CURL_D << "getFtpDetails() c:" << c->getDirPath()
                  << " serv:" << d.serv
                  << " path:" << d.path
                  << endl;
        
        return d;
    }


    /************************************************************/
    /************************************************************/
    /************************************************************/

    /*
     * set global options from dotfile configurations.
     */
    class FERRISEXP_DLLLOCAL FerrisCurlGlobalClass
    {
        FerrisCurlGlobalClass( const FerrisCurlGlobalClass& );
        
    public:
        FerrisCurlGlobalClass()
            {
                long iflags = CURL_GLOBAL_NOTHING;

                string useSSL = getEDBString( FDB_GENERAL, "curl-use-ssl", "0" );
                if( isTrue( useSSL ) )
                {
                    iflags |= CURL_GLOBAL_SSL;
                }
                
                CURLcode r = curl_global_init( iflags );
                switch( r )
                {
                case CURLE_OK: break;
                default:
                    fh_stringstream ss;
                    ss << "curl_global_init() failed. iflags:" << iflags
                       << " return code:" << r;
                    Throw_CurlStaticInitFailed( tostr(ss), 0 );
                }
            }
        ~FerrisCurlGlobalClass()
            {
                curl_global_cleanup();
            }
    };
    typedef Loki::SingletonHolder<FerrisCurlGlobalClass> FerrisCurlGlobal;
    
    /************************************************************/
    /************************************************************/
    /************************************************************/

    size_t FerrisCurlState_hdr_cb( void *ptr, size_t size, size_t nmemb, void *stream);
    int    FerrisCurlState_progress_cb(void *clientp, size_t dltotal, size_t dlnow,
                                       size_t ultotal, size_t ulnow);
    size_t FerrisCurlState_write_cb( void *ptr, size_t size, size_t nmemb, void *stream);
    size_t FerrisCurlState_read_cb ( void *ptr, size_t size, size_t nmemb, void *stream);

    /************************************************************/
    /************************************************************/
    /************************************************************/
    
    typedef Loki::SmartPtr<
        FerrisCurlState,
        Loki::RefCounted,
        Loki::DisallowConversion,
        Loki::AssertCheck,
        Loki::DefaultSPStorage > fh_curlstate;
    class FERRISEXP_DLLLOCAL FerrisCurlState
        :
        public Handlable
    {
    public:
        typedef list< string > headers_t;

        typedef SigC::Signal2< void, FerrisCurlState&, fh_iostream& > WriteSig_t;
        WriteSig_t  WriteSig;
        WriteSig_t& getWriteSig()
            {
                return WriteSig;
            }

        typedef SigC::Signal2< void, FerrisCurlState*, std::string > HeaderSig_t;
        HeaderSig_t HeaderSig;
        HeaderSig_t& getHeaderSig()
            {
                return HeaderSig;
            }
        
        streamsize BytesReadCount;
        void clearBytesReadCount()
            {
                BytesReadCount = 0;
            }
        streamsize getBytesReadCount()
            {
                return BytesReadCount;
            }
        
        
        typedef SigC::Signal2< streamsize, FerrisCurlState&, fh_iostream > ReadSig_t;
        ReadSig_t  ReadSig;
        ReadSig_t& getReadSig()
            {
                return ReadSig;
            }
        
    private:
        
        friend class FerrisCurlClass;
        friend size_t FerrisCurlState_write_cb( void *ptr, size_t size, size_t nmemb, void * );
        friend size_t FerrisCurlState_read_cb ( void *ptr, size_t size, size_t nmemb, void * );
        


        // data is stored in here in the write callback and a WriteSig is emitted
        // to let the client slurp the data out.
        fh_iostream writeStream;

        // data is stored here on its way to the network
        fh_iostream readStream;

        // servername and path
        FtpDetails theFtpDetails;

        // URL for the server ftp://serv/path/goes/here
        string earl;

        // For giving a custom ftp/http command
        string customRequest;

        // Some FTP commands to run
        struct curl_slist* FtpCmdPostList;
        struct curl_slist* FtpCmdList;

        typedef list< string > stringcache_t;
        stringcache_t stringcache;

        typedef map< CURLoption, void* > curlops_t;
        curlops_t curlops;

        headers_t headers;
        
        void setup( CURL* c )
            {
                writeStream.seekg(0);
                writeStream.seekp(0);
                writeStream.clear();

                readStream.seekg(0);
                readStream.seekp(0);
                readStream.clear();

                clearBytesReadCount();
                
                char errbuf[CURL_ERROR_SIZE];
                errbuf[0] = 0;
                curl_easy_setopt( c, CURLOPT_ERRORBUFFER, errbuf );

                LG_CURL_D << "setup() url:" << earl << endl;
                
                /* URL is set first so that it can be overriden */
                curl_easy_setopt( c, CURLOPT_URL, earl.c_str() );

                for( curlops_t::iterator iter = curlops.begin();
                     iter != curlops.end(); ++iter )
                {
                    CURLcode r = curl_easy_setopt( c, iter->first, iter->second );
                    if( r != CURLE_OK )
                    {
                        fh_stringstream ss;
                        ss << "curl_easy_setopt() failed:" << errbuf
                           << " return code:" << r;
                        Throw_CurlStateError( tostr(ss), 0 );
                    }
                }

                curl_easy_setopt( c, CURLOPT_POSTQUOTE, FtpCmdPostList );
                curl_easy_setopt( c, CURLOPT_QUOTE,     FtpCmdList );


                if( customRequest.length() )
                {
                    curl_easy_setopt( c, CURLOPT_CUSTOMREQUEST, customRequest.c_str() );
                }
                else
                {
                    curl_easy_setopt( c, CURLOPT_CUSTOMREQUEST, 0 );
                }

                LG_CURL_D << "state::setup() setting mtime to:" << mtime <<  endl;

                curl_easy_setopt( c, CURLOPT_TIMECONDITION, CURL_TIMECOND_IFMODSINCE );
                curl_easy_setopt( c, CURLOPT_TIMEVALUE, mtime );
                
            }

        void setEDBint( CURLoption op, const string& eaname, long d )
            {
                string s = getEDBString( FDB_GENERAL, eaname, tostr(d) );
                if( s.length() )
                {
                    long pnum = -1;
                    fh_stringstream ss;
                    ss << s;
                    ss >> pnum;

                    if( pnum != -1 )
                    {
                        set( op, pnum );
                    }
                }
                
            }

        void clearSList( struct curl_slist*& l )
            {
                curl_slist_free_all( l );
                l = 0;
            }

        void appendSList( struct curl_slist*& l, const string& s )
            {
                // append() strdups for us
                l = curl_slist_append( l, s.c_str());
            }
        

        time_t mtime;
        
    public:

        void setDownloadIfMTimeSince( time_t v )
            {
                mtime = v;
            }
        

        void setCustomRequest( const string& s )
            {
                customRequest = s;
            }
        
        void clearFtpCmdPost()
            {
                clearSList( FtpCmdPostList );
            }

        void appendFtpCmdPost( const string& s )
            {
                appendSList( FtpCmdPostList, s );
            }

        void clearFtpCmd()
            {
                clearSList( FtpCmdList );
            }

        void appendFtpCmd( const string& s )
            {
                appendSList( FtpCmdList, s );
            }
        
        void appendHeader( const string& s )
            {
                headers.push_back( s );
            }

        headers_t& getHeaders()
            {
                return headers;
            }
        
        
        FerrisCurlState()
            :
            FtpCmdPostList(0),
            FtpCmdList(0),
            writeStream( fh_stringstream() ),
            readStream( fh_stringstream() ),
            mtime( 0 )
            {
                if( isTrue( getEDBString( FDB_GENERAL, "curl-use-proxy", "" )))
                {
                    string proxyname = getEDBString( FDB_GENERAL, "curl-use-proxy-name", "" );
                    if( proxyname.length() )
                    {
                        set( CURLOPT_PROXY, proxyname );
                    }

                    setEDBint( CURLOPT_PROXYPORT, "curl-use-proxy-port", -1 );
                    
                    string proxyuserpass =
                        getEDBString( FDB_GENERAL, "curl-use-proxy-userpass", "" );
                    
                    if( !proxyuserpass.empty() )
                    {
                        set( CURLOPT_PROXYUSERPWD, proxyuserpass );
                    }

                    if( isTrue( getEDBString( FDB_GENERAL, "curl-use-proxy-tunnel", "0" )))
                    {
                        set( CURLOPT_HTTPPROXYTUNNEL, 1 );
                    }
                }
                
                ////////////////////////////////////////////////////////////
                ////////////////////////////////////////////////////////////
                ////////////////////////////////////////////////////////////
                
                {
                    string t = getEDBString( FDB_GENERAL, "curl-transfer-timeout", "15" );
                    if( t.empty() )
                        t = "15";
                    if( int n = toint( t))
                    {
                        set( CURLOPT_TIMEOUT, n );
                    }
                }
                
                

                if( isTrue( getEDBString( FDB_GENERAL, "curl-verbose", "1" )))
                {
                    set( CURLOPT_VERBOSE, 1 );
                }

                if( isTrue( getEDBString( FDB_GENERAL, "curl-use-netrc", "0" )))
                {
                    set( CURLOPT_NETRC, 1 );
                }

                if( isTrue( getEDBString( FDB_GENERAL, "curl-follow", "0" )))
                {
                    set( CURLOPT_FOLLOWLOCATION, 1 );
                }
                setEDBint( CURLOPT_MAXREDIRS, "curl-max-redirects", 16 );

                string agentname = getEDBString( FDB_GENERAL, "curl-agent-name", "0" );
                if( agentname.length() )
                {
                    set( CURLOPT_USERAGENT, agentname );
                }

                setEDBint( CURLOPT_SSLVERSION, "curl-use-ssl-version-number", -1 );

                set( CURLOPT_HEADERFUNCTION, (void*)FerrisCurlState_hdr_cb );
                set( CURLOPT_WRITEHEADER, (void*)this );

                set( CURLOPT_PROGRESSDATA, (void*)this );
                set( CURLOPT_PROGRESSFUNCTION, (void*)FerrisCurlState_progress_cb );

                set( CURLOPT_FILE, (void*)this );
                set( CURLOPT_WRITEFUNCTION, (void*)FerrisCurlState_write_cb );

                set( CURLOPT_INFILE, (void*)this );
                set( CURLOPT_READFUNCTION, (void*)FerrisCurlState_read_cb );
                
                string ifname = getEDBString( FDB_GENERAL, "curl-client-interface", "" );
                if( ifname.length() )
                {
                    set( CURLOPT_INTERFACE, ifname );
                }
            }

        ~FerrisCurlState()
            {
                clearFtpCmdPost();
                clearFtpCmd();
            }

        void set( CURLoption op, const string& s )
            {
                stringcache.push_back( s );
                curlops[ op ] = (void*)stringcache.back().c_str();
            }
        
        void set( CURLoption op, void* p )
            {
                curlops[ op ] = p;
            }
        
        void set( CURLoption op, long l )
            {
                curlops[ op ] = (void*)l;
            }

        time_t getinfo_FILETIME();
        streamsize getinfo_CONTENT_LENGTH_DOWNLOAD();

        void setFtpDetails( const FtpDetails& f )
            {
                theFtpDetails = f;
            }

        FtpDetails getFtpDetails()
            {
                return theFtpDetails;
            }

        bool isHTTP()
            {
                LG_CURL_D << "state::isHTTP() serv:" << theFtpDetails.serv
                          << " earl:" << earl
                          << " ret:" << starts_with( earl, "http:" )
                          << endl;
                return starts_with( earl, "http:" );
            }
        
        void setupURL( const string& meth = "ftp://", bool getAsFile = false )
            {
                clearFtpCmd();
                clearFtpCmdPost();
                set( CURLOPT_NOBODY, (long int)0 );

                fh_stringstream ss;
                ss << meth << theFtpDetails.serv;
//                ss << "/" << theFtpDetails.path;
                {
                    // ensure that there is only one slash at the root of the path
                    string p = "/";
                    p+=theFtpDetails.path;
                    PrefixTrimmer trimmer;
                    trimmer.push_back( "/" );
                    ss << "/" << trimmer( p );
                }
                if( !getAsFile )
                    ss << "/";

                earl = tostr(ss);
                LG_CURL_D << "setupURL() getAsFile:" << getAsFile << " earl:" << earl << endl;
                LG_CURL_D << "setupURL() getAsFile:" << getAsFile
                          << " earl:" << earl
                          << " meth:" << meth
                          << " theFtpDetails.serv:" << theFtpDetails.serv
                          << " theFtpDetails.path:" << theFtpDetails.path
                          << endl;

                if( !isHTTP() )
                {
                    // Seems that using a proxy doesn't mix FTP
                    set( CURLOPT_PROXYUSERPWD, (char*)0 );
                    set( CURLOPT_HTTPPROXYTUNNEL, (char*)0 );
                    set( CURLOPT_PROXY, (void*)0 );
                }
                
            }

        void print()
            {
                LG_CURL_D << "FerrisCurlState. serv:" << theFtpDetails.serv << endl;
            }
    };

    size_t FerrisCurlState_hdr_cb( void *ptr, size_t size, size_t nmemb, void *user_data)
    {
//         char null = '\0';
//         FerrisCurlState* s = (FerrisCurlState*)user_data;
//         stringstream ss;
//         ss.write( (const char*)ptr, size*nmemb );
// //        ss.write( &null, 1 );
//         s->appendHeader( ss.str() );
//         s->getHeaderSig().emit( s, tostr(ss) );
        return size*nmemb;
    }
    
    int FerrisCurlState_progress_cb(void *user_data, size_t dltotal, size_t dlnow,
                                    size_t ultotal, size_t ulnow)
    {
        FerrisCurlState* s = (FerrisCurlState*)user_data;
        return 0;
    }
    
    size_t FerrisCurlState_write_cb( void *ptr, size_t size, size_t nmemb, void *user_data)
    {
        FerrisCurlState* s = (FerrisCurlState*)user_data;
        size_t totalsize = size * nmemb;

        LG_CURL_D << "FerrisCurlState_write_cb() totalsize:" << totalsize << endl;
        LG_CURL_D << "FerrisCurlState_write_cb() ptr:" << (const char*)ptr << endl;
        s->writeStream.clear();
        s->writeStream.write( (const char*)ptr, totalsize );
        s->getWriteSig().emit( *s, s->writeStream );
        return totalsize;
    }

    size_t FerrisCurlState_read_cb ( void *ptr, size_t size, size_t nmemb, void *user_data )
    {
        FerrisCurlState* s = (FerrisCurlState*)user_data;
        size_t totalsize = size * nmemb;

        LG_CURL_D << "FerrisCurlState_read_cb() totalsize:" << totalsize << endl;

        /*
         * If we have the ability to send more data then give the user the chance to
         * send more.
         */
        streamsize AvailableDataSize = s->readStream->tellp() - s->readStream->tellg();
        LG_CURL_D << "FerrisCurlState_read_cb()"
                  << " inital AvailableDataSize:" << AvailableDataSize
                  << endl;
        if( AvailableDataSize < totalsize )
        {
            AvailableDataSize += s->getReadSig().emit( *s, s->readStream );
            LG_CURL_D << "FerrisCurlState_read_cb()"
                      << " after getReadSig() fired AvailableDataSize:" << AvailableDataSize
                      << endl;
        }
        

//         /* Figure out how much data is available if the user doesn't know */
//         if( !AvailableDataSize )
//         {
//             AvailableDataSize = s->readStream->tellp() - s->readStream->tellg();
//         }

        /*
         * We assume that if there was an error then the signal handler
         * would have thrown an exception or done something about the
         * error.
         */
        if( !AvailableDataSize )
        {
            return static_cast<size_t>(0);
        }
        
        if( !s->readStream )
        {
            LG_CURL_D << "FerrisCurlState_read_cb() read failed totalsize:" << totalsize
                      << " AvailableDataSize:" << AvailableDataSize
                      << endl;
            return static_cast<size_t>(-1);
        }

        /*
         * Clamp the amount of data that we are going to send to me <= what curl
         * can handle at this time
         */
        if( AvailableDataSize < totalsize )
            totalsize = AvailableDataSize;

        s->readStream.read( (char*)ptr, totalsize );
        totalsize = static_cast<size_t>( s->readStream.gcount() );
        s->BytesReadCount += totalsize;

        return totalsize ? totalsize : static_cast<size_t>(-1);
    }

    
    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/
    
    
    class FERRISEXP_DLLLOCAL FerrisCurlClass 
    {
        CURL* c;
        
    public:
        
        FerrisCurlClass();
        ~FerrisCurlClass();

        CURL* getCurl();
        
        fh_curlstate operator()( curlFtpContext* ctx, fh_curlstate state );
    };
    typedef Loki::SingletonHolder<FerrisCurlClass> FerrisCurl;

    
//     fh_curlstate CurlPerform( fh_curlstate s )
//     {
//         return FerrisCurl::Instance()(s);
// //         FerrisCurlClass& c = FerrisCurl::Instance();
// //         return c(s);
//     }

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    fh_curlstate getState( fh_context c )
    {
        typedef map< string, fh_curlstate > states_t;
        states_t states;

        LG_CURL_D << "getState() c:" << c->getDirPath() << endl;
        
        const FtpDetails& fdt    = getFtpDetails( c );
        const string& servername = fdt.serv;
        const string& path       = fdt.path;

        LG_CURL_D << "getState() c:" << c->getURL()
                  << " serv:" << servername
                  << " path:" << path
                  << endl;
        
        
        states_t::iterator iter = states.find( servername );
        if( iter != states.end() )
        {
            LG_CURL_D << "getState() connection cached c:" << c->getDirPath() << endl;
            fh_curlstate st = GetImpl(iter->second);
            st->setFtpDetails( fdt );

            LG_CURL_D << "getState(2) c:" << c->getURL()
                      << " serv:" << st->getFtpDetails().serv
                      << " path:" << st->getFtpDetails().path
                      << endl;
            
            return st;
        }

        fh_curlstate st = new FerrisCurlState();
        LG_CURL_D << "getState() new connection created c:" << c->getDirPath() << endl;
        states[servername] = st;
        st->setFtpDetails( fdt );

//         LG_CURL_D << "getState(3) c:" << c->getURL()
//                   << " serv:" << st->getFtpDetails().serv
//                   << " path:" << st->getFtpDetails().path
//                   << endl;
//         st->print();
            
        
        return st;
    }
    
    
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

    
    class FERRISEXP_CTXPLUGIN curlFtpContext
        :
        public StateLessEAHolder< curlFtpContext, FakeInternalContext >
    {
        typedef curlFtpContext _Self;
        typedef StateLessEAHolder< curlFtpContext, FakeInternalContext > _Base;
        friend fh_istream SL_getMTimeRawStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom );
        friend fh_istream SL_getSizeStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom );
//         friend fh_istream SL_getDownloadIfMTimeSinceIStream( curlFtpContext* c,
//                                                              const std::string& rdn,
//                                                              EA_Atom* atom );
//         friend fh_iostream SL_getDownloadIfMTimeSinceIOStream( curlFtpContext* c,
//                                                                const std::string& rdn,
//                                                                EA_Atom* atom );
//         friend void SL_downloadIfMTimeSinceClosed( curlFtpContext* c,
//                                                       const std::string& rdn,
//                                                       EA_Atom* atom, fh_istream ss );

        friend fh_istream SL_getUserOwnerNumberStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom );
        friend fh_istream SL_getGroupOwnerNumberStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom );

        virtual fh_context priv_getSubContext( const std::string& rdn )
            throw( NoSuchSubContext );
        
    protected:

        virtual bool isRemote()
            {
                return true;
            }
        
        string getRecommendedEA()
            {
                return "protection-ls,size-human-readable,mtime-display,name";
            }
        
        virtual bool VetoEA()
            {
                return true;
            }

        fh_context getRootOfContext( fh_context c )
            {
                while( c->isParentBound() )
                {
                    c = c->getParent();
                }
                return c;
            }

        string getMethod()
            {
                fh_context rc = getRootOfContext( this );
                string rccName = RootContextFactory::getRootContextClassName(rc);
                LG_CURL_D << "getMethod() rccName:" << rccName << endl;
                return rccName + "://";
                
//                 string url = getURL();
//                 if( url.find( "ftp://" ) != string::npos )
//                     return "ftp://";
//                 if( url.find( "http://" ) != string::npos )
//                     return "http://";
//                 return "ftp://";
            }
        

        ferris_ios::openmode
        getSupportedOpenModes()
            {
                return ios::in | ios::out | ios::binary;
            }

        typedef list< string > months_t;
        months_t& getMonths();
        
        void priv_read_addline( const string& linestr );
        void priv_read_wrh( FerrisCurlState& state, fh_iostream& ss );
        virtual void priv_read();
        virtual void createStateLessAttributes( bool force = false );

        virtual fh_iostream  getRealStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   CanNotGetStream,
                   exception);
        void getRealStream_wrh( FerrisCurlState& state,    fh_iostream& ss,
                                bool headerSignalFired,
                                stringset_t* strset,
                                fh_stringstream );
        void getRealStream_header_wrh( FerrisCurlState* s, string hdr, stringset_t* strset  );

        
        virtual fh_istream  priv_getIStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   CanNotGetStream,
                   exception);
        virtual fh_iostream priv_getIOStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   AttributeNotWritable,
                   CanNotGetStream,
                   exception);

        streamsize priv_getIOStream_IOStreamClosed_readh(
            FerrisCurlState& state, fh_iostream ss, fh_istream );
        void priv_getIOStream_IOStreamClosed( fh_istream& ss, std::streamsize tellp );

        void priv_createSubContext_writeh( FerrisCurlState& state, fh_iostream& ss );
        virtual fh_context
        priv_createSubContext( const string& rdn,  fh_context md )
            throw( FerrisCreateSubContextFailed, FerrisCreateSubContextNotSupported );
        
        time_t mtime;
        bool m_isDir;
        streamsize theSize;
//         long m_downloadIfMTimeSince;
        
    public:

        curlFtpContext( Context* parent, const std::string& rdn )
            :
            theSize(0),
            mtime(0),
            m_isDir( false )
//             m_downloadIfMTimeSince( 0 )
            {
                setContext( parent, rdn );
                createStateLessAttributes();
                addAttribute( EAN_IS_REMOTE, "1", XSD_BASIC_BOOL );
            }

        fh_curlstate CurlPerform( fh_curlstate s )
            {
                return FerrisCurl::Instance()( this, s );
            }
        
        void setMTime( time_t t )
            {
                mtime = t;
            }
        void setIsDir( bool v )
            {
                m_isDir = v;
            }
        virtual bool isDir()
            {
                return m_isDir;
            }
        
        

//         long getDownloadIfMTimeSince()
//             {
//                 return m_downloadIfMTimeSince;
//             }

//         void setDownloadIfMTimeSince( long x )
//             {
//                 m_downloadIfMTimeSince = x;
//             }
        
        streamsize getSize()
            {
                return theSize;
            }
        
        void setSize( string s )
            {
                LG_CURL_D << "SETTING theSize to:" << s << endl;
                theSize = toType< streamsize >(s);
                
//                 fh_stringstream ss;
//                 ss << s;
//                 ss >> theSize;
            }

        static const std::string CreateObjectType_k;
        static const std::string CreateObjectType_v_Dir;
        static const std::string CreateObjectType_v_File;
        static const std::string CreateObjectType_k_FileForceCreate;
    };

    /*
     * Allow network hosts to be discovered at run time.
     */
    class FERRISEXP_CTXPLUGIN curlFtpRootContext
        :
        public networkRootContext<curlFtpContext>
    {
        typedef curlFtpRootContext                  _Self;
        typedef networkRootContext<curlFtpContext>  _Base;
    public:
        
        curlFtpRootContext( Context* parent, const std::string& rdn )
            :
            networkRootContext<curlFtpContext>( parent, rdn, true )
            {}
        
        virtual ~curlFtpRootContext()
            {}
    };
    
    

    /*******************************************************************************/
    /*******************************************************************************/
    /*******************************************************************************/
    
    const string curlFtpContext::CreateObjectType_k      = "CreateObjectType";
    const string curlFtpContext::CreateObjectType_v_Dir  = "Dir";
    const string curlFtpContext::CreateObjectType_v_File = "File";
    const string curlFtpContext::CreateObjectType_k_FileForceCreate = "FileForceCreate";
    
    fh_istream
    SL_getMTimeRawStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom )
    {
        c->ensureUpdateMetaDataCalled();
        time_t t = c->mtime;

        LG_CURL_D << "SL_getMTimeRawStream() initial mtime:" << t << endl;
        
        if( !t )
        {
            try
            {
                LG_CURL_D << "SL_getMTimeRawStream() getting remote mtime" << endl;
                fh_curlstate st = getState( c );
                st->setupURL( c->getMethod(), true );
                if( st->isHTTP() )
                {
                    st->set( CURLOPT_NOBODY,   1 );
                    st->set( CURLOPT_FILETIME, 1 );
                    c->CurlPerform( st );
                    t = st->getinfo_FILETIME();
                    LG_CURL_D << "SL_getMTimeRawStream() getting remote mtime t:" << t << endl;
                }
                else
                {
                    LG_CURL_D << "SL_getMTimeRawStream() ftp request... c:" << c->getURL() << endl;
                    if( c->isParentBound() )
                        c->getParent()->read();
                    t = c->mtime;
                }
            }
            catch( exception& e )
            {
                LG_CURL_W << "SL_getMTimeRawStream() e:" << e.what() << endl;
                throw;
            }
        }
        
        fh_stringstream ss;
        ss << t;
        return ss;
    }

    fh_istream
    SL_getSizeStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom )
    {
        LG_CURL_D << "SL_getSizeStream() called c:" << c->getURL() << endl;
        
        c->ensureUpdateMetaDataCalled();
        streamsize sz = c->getSize();
        LG_CURL_D << "SL_getSizeStream() top sz:" << sz << " c:" << c->getURL() << endl;

        if( !sz )
        {
            try
            {
                LG_CURL_D << "SL_getSizeStream() getting remote sz c:" << c->getURL() << endl;
                fh_curlstate st = getState( c );
                st->setupURL( c->getMethod(), true );
                if( st->isHTTP() )
                {
                    LG_CURL_D << "SL_getSizeStream() http request... c:" << c->getURL() << endl;
                    st->set( CURLOPT_NOBODY,   1 );
                    c->CurlPerform( st );
                    sz = st->getinfo_CONTENT_LENGTH_DOWNLOAD();
                }
                else
                {
                    LG_CURL_D << "SL_getSizeStream() ftp request... c:" << c->getURL() << endl;
                    if( c->isParentBound() )
                        c->getParent()->read();
                    sz = c->getSize();
                }

                LG_CURL_D << "SL_getSizeStream() getting remote sz size is:" << sz << endl;
            }
            catch( exception& e )
            {
                LG_CURL_W << "SL_getSizeStream() e:" << e.what() << endl;
                throw;
            }
        }
        
        fh_stringstream ss;
        ss << sz;
        return ss;
    }

    fh_istream
    SL_getUserOwnerNumberStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom )
    {
        fh_stringstream ss;
        ss << Shell::getUserID( FERRIS_USER_NAME_NOBODY );
        return ss;
    }

    fh_istream
    SL_getGroupOwnerNumberStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom )
    {
        fh_stringstream ss;
        ss << Shell::getGroupID( FERRIS_USER_NAME_NOBODY );
        return ss;
    }
    
//     fh_istream
//     SL_getDownloadIfMTimeSinceIStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom )
//     {
//         c->ensureUpdateMetaDataCalled();
//         fh_stringstream ss;
//         ss << c->getDownloadIfMTimeSince();
//         cerr << "SL_getDownloadIfMTimeSinceIStream() t:" << c->getDownloadIfMTimeSince() << endl;
//         return ss;
//     }

//     fh_iostream
//     SL_getDownloadIfMTimeSinceIOStream( curlFtpContext* c, const std::string& rdn, EA_Atom* atom )
//     {
//         c->ensureUpdateMetaDataCalled();
//         fh_stringstream ss;
//         ss << c->getDownloadIfMTimeSince();
//         cerr << "SL_getDownloadIfMTimeSinceIOStream() t:" << c->getDownloadIfMTimeSince() << endl;
//         return ss;
//     }

//     void
//     SL_downloadIfMTimeSinceClosed( curlFtpContext* c, const std::string& rdn, EA_Atom* atom, fh_istream ss )
//     {
//         time_t x = 0;
//         ss >> x;
//         cerr << "SL_downloadIfMTimeSinceClosed() c:" << c->getURL()
//              << " x:" << x
//              << " x.str:" << Time::toTimeString( x )
//              << endl;
//         c->setDownloadIfMTimeSince( x );
//     }
    
    
    void
    curlFtpContext::createStateLessAttributes( bool force )
    {
        static Util::SingleShot virgin;
        if( virgin() )
        {
            LG_CURL_D << "curlFtpContext::createStateLessAttributes()" << endl;
            tryAddStateLessAttribute( "mtime", SL_getMTimeRawStream, FXD_UNIXEPOCH_T );
            tryAddStateLessAttribute( "size",  SL_getSizeStream,     FXD_FILESIZE );
            tryAddStateLessAttribute( "user-owner-number",
                                      SL_getUserOwnerNumberStream,
                                      FXD_UID_T );
            tryAddStateLessAttribute( "group-owner-number",
                                      SL_getGroupOwnerNumberStream,
                                      FXD_GID_T );
            
//             tryAddStateLessAttribute( "download-if-mtime-since",
//                                       SL_getDownloadIfMTimeSinceIStream,
//                                       SL_getDownloadIfMTimeSinceIOStream,
//                                       SL_downloadIfMTimeSinceClosed,
//                                       FXD_UNIXEPOCH_T );
            
            _Base::createStateLessAttributes( true );
            supplementStateLessAttributes( true );
        }
    }
    
    /**
     * Get the next token delimited by delim
     */
    string getToken( fh_iostream& ss, char delim = ' ' )
    {
        if( !ss ) return "";
        
        string ret;
        char c = delim;

        while( ss >> c && c == delim )
        {
        }
        getline(ss, ret, delim );
        if( !ss ) ret = "";
        ret.insert( ret.begin(), c );
        
        return ret;
    }


    //
    // lines like
    // drwxr-xr-x   2 root     50            114 Nov 20 17:08 pub
    // drwxr-xr-x   2 root     114 Nov 20 17:08 pub
    //
    void
    curlFtpContext::priv_read_addline( const string& linestr )
    {
        fh_stringstream linestream;
        linestream << linestr;

        string protection_ls    = "protection-ls";
        string hard_link_count  = "hard-link-count";
        string user_owner_name  = "user-owner-name";
        string group_owner_name = "group-owner-name";
        string size             = "size";
        string month            = "month";
        string dayofmonth       = "dayofmonth";
        string time_or_year     = "time-or-year";
        string year             = "year";
        string time             = "time";

        typedef map< string, string > ea_t;
        ea_t ea;

        ea[ protection_ls ]    = getToken( linestream );
        ea[ hard_link_count ]  = getToken( linestream );
        ea[ user_owner_name ]  = getToken( linestream );
        ea[ group_owner_name ] = getToken( linestream );
        ea[ size ]             = getToken( linestream );
        if( find( getMonths().begin(), getMonths().end(), ea[ size ] ) != getMonths().end() )
        {
            // lag: drwxr-xr-x   2 root     50            114 Nov 20 17:08 pub
            // la : drwxr-xr-x   2 root     114 Nov 20 17:08 pub
            //
            // If we only have la mode then we don't know the group, and have read
            // the size as the group and the month as the size.
            //
            ea[ month ] = ea[ size ];
            ea[ size ]  = ea[ group_owner_name ];
            ea.erase( group_owner_name );
        }
        else
        {
            ea[ month ]            = getToken( linestream );
        }
        
        ea[ dayofmonth ]       = getToken( linestream );
        ea[ time_or_year ]     = getToken( linestream );
        string rdn;
        getline( linestream, rdn );
        PostfixTrimmer trimmer;
        trimmer.push_back( "\r" );
        trimmer.push_back( "\n" );
        rdn = trimmer( rdn );
        {
            int arrowPos = rdn.find("->");
            if( string::npos != arrowPos )
            {
                rdn = rdn.substr( 0, arrowPos );
                PostfixTrimmer trimmer;
                trimmer.push_back( " " );
                trimmer.push_back( "\t" );
                rdn = trimmer( rdn );
            }
        }
        LG_CURL_D << "curlFtpContext::priv_read_wrh() "
                  << " rdn :-->:" << rdn << ":<--: "
                  << " linestr:" << linestr
                  << endl;

        if( rdn == "." || rdn == ".." )
            return;
        
        if( !rdn.length() )
        {
            LG_CURL_ER << "curlFtpContext::priv_read_addline()"
                       << " can not determine file name" << endl;
            return;
        }

        /*
         * year_or_time is either the year of mod if it is not the current year,
         * or the mtime if the year is the curent year.
         */
        if( ea[ time_or_year ].find(":") == string::npos )
        {
            ea[ year ] = ea [ time_or_year ];
            ea.erase( time_or_year );
        }
        else
        {
            time_t t = ::time(0);
            struct tm* tp = localtime( &t );
            ea[ year ] = tostr( tp->tm_year + 1900 );
            ea[ time ] = ea[ time_or_year ];
            ea.erase( time_or_year );
        }

        curlFtpContext* child = 0;
        child = priv_ensureSubContext( rdn, child );

        for( ea_t::iterator iter = ea.begin(); iter != ea.end(); ++iter )
        {
            LG_CURL_D << "k:" << iter->first << " v:" << iter->second << endl;
            child->addAttribute( iter->first, iter->second, XSD_BASIC_STRING );
        }
        LG_CURL_D << "curlFtpContext::priv_read_wrh() done printing ea "
                  << " for rdn:" << rdn
                  << endl;

        child->setSize( ea[ size ] );

        /* Make a mtime from the given data */
        if( ea.end() != ea.find( dayofmonth )
            && ea.end() != ea.find( month )
            && ea.end() != ea.find( year ) )
        {
            GDate gd;
            g_date_clear( &gd, 1 );
            fh_stringstream dmyss;
            
            dmyss << ea[ dayofmonth ] << " " << ea[ month ] << " " <<  ea[ year ];
            string dmystr = tostr(dmyss);
            g_date_set_parse( &gd, dmystr.c_str() );
            if( !g_date_valid( &gd ) )
            {
                LG_CURL_D << "invalid mtime for rdn:" << rdn << endl;
            }
            else
            {
                struct tm tm;
                g_date_to_struct_tm( &gd, &tm );
                fh_stringstream timess;
                timess << ea[ time ];
                string s;
                tm.tm_sec = 0;
                if( getline( timess, s, ':' ) )
                {
                    tm.tm_hour = toint(s);
                    if( getline( timess, s ) )
                    {
                        tm.tm_min = toint(s);
                    }
                }
                time_t mtime = mktime( &tm );
                LG_CURL_D << "curlFtpContext::priv_read_wrh()"
                          << " mtime:" << ctime( &mtime ) << endl;
                        
                /* Stateless EA value */
                child->setMTime( mtime );
            }
        }

        if( ea.find( protection_ls ) != ea.end() )
        {
            string p = ea[ protection_ls ];
            if( !p.empty() )
                child->setIsDir( p[0]=='d' );
        }
        
        LG_CURL_D << "curlFtpContext::priv_read_wrh() done "
                  << " for rdn:" << rdn
                  << endl;
    }
    
    curlFtpContext::months_t&
    curlFtpContext::getMonths()
    {
        static bool virgin = true;
        static months_t months;

        if( virgin )
        {
            virgin = false;

            months.push_back( "Jan" );
            months.push_back( "Feb" );
            months.push_back( "Mar" );
            months.push_back( "Apr" );
            months.push_back( "May" );
            months.push_back( "Jun" );
            months.push_back( "Jul" );
            months.push_back( "Aug" );
            months.push_back( "Sep" );
            months.push_back( "Oct" );
            months.push_back( "Nov" );
            months.push_back( "Dec" );
        }
        return months;
    }
    

    
    void
    curlFtpContext::priv_read_wrh( FerrisCurlState& state, fh_iostream& ss )
    {
        try
        {
            string s;
        
            LG_CURL_D << "priv_read_wrh( TOP ) s:" << s << endl;

            while( getline( ss, s ) )
            {
                LG_CURL_D << "priv_read_wrh() s:" << s
                          << endl;

                if( starts_with( s, "total" ))
                {
                    continue;
                }

                if( ss.eof() )
                {
                    LG_CURL_D << "EOF bit set on s:" << s << endl;
                    ss.clear();
                    ss.seekg(  -1 * s.length(), ios::end );
//                    ss << s;
                    return;
                }

                LG_CURL_D << "priv_read_wrh() s:" << s << endl;
                priv_read_addline( s );
            }
        }
        catch( exception& e )
        {
            LG_CURL_D << "curlFtpContext::priv_read_wrh()"
                      << " e:" << e.what()
                      << endl;
            throw e;
        }
        
    }
    
    fh_context
    curlFtpContext::priv_getSubContext( const std::string& rdn )
        throw( NoSuchSubContext )
    {
        if( getMethod() == "http://" )
        {
            curlFtpContext* child = 0;
            child = priv_ensureSubContext( rdn, child );
            return child;
        }
        return _Base::priv_getSubContext( rdn );
    }
    

    void
    curlFtpContext::priv_read()
    {
        LG_CURL_D << "curlFtpContext::priv_read() path:" << getDirPath() << endl;

        EnsureStartStopReadingIsFiredRAII _raii1( this );
        clearContext();
        LG_CURL_D << "priv_read(top) path:" << getDirPath() << endl;

        try
        {
            fh_curlstate st = getState( this );
            st->setupURL( getMethod() );
            if( getMethod() == "ftp://" )
            {
                st->setCustomRequest( "LIST -lga" );
            }
            if( st->isHTTP() )
            {
                stringstream ss;
                ss << "Can not read HTTP directories. url:" << getURL() << endl;
                Throw_FerrisNotReadableAsContext( tostr(ss), this );
            }
            // Seems that using a proxy doesn't mix with setCustomRequest
            // as its sent as an HTTP request NOT FTP when using the proxy!!
            st->set( CURLOPT_PROXY, (void*)0 );
            
            LG_CURL_D << "priv_read(2) about to perform path:" << getDirPath() << endl;

            SigC::Connection con = st->getWriteSig().connect(
                SigC::slot( *this,
                      &curlFtpContext::priv_read_wrh ));
            if( st->isHTTP() )
                con.disconnect();

            CurlPerform( st );

            if( !st->isHTTP() )
                con.disconnect();

            LG_CURL_D << "priv_read(3) perform done path:" << getDirPath() << endl;
        }
        catch( CanNotReadContext& e )
        {
            throw;
        }
        catch( CurlPerformFailed& e )
        {
            LG_CURL_D << "priv_read(e) error path:" << getDirPath()
                      << " e:" << e.what()
                      << endl;
            //
            // FTP returns 550 Not a directory. which curl translates to 9
            // ACCESS_DENIED
            //
            //
            // why did I want to ignore this?
            //
            if( 0 && e.getResultCode() == CURLE_FTP_ACCESS_DENIED )
            {
                LG_CURL_D << "e.getResultCode() == CURLE_FTP_ACCESS_DENIED" << endl;
            }
            else
            {
                LG_CURL_D << "rethrowing e:" << e.what() << endl;
                throw e;
            }
            
        }
        
        
        LG_CURL_D << "priv_read(done) path:" << getDirPath() << endl;
    }


    fh_istream
    curlFtpContext::priv_getIStream( ferris_ios::openmode m )
        throw (FerrisParentNotSetError,
               CanNotGetStream,
               exception)
    {
        LG_CURL_D << "getting I stream..." << endl;
        return getRealStream( m );
    }

    streamsize
    curlFtpContext::priv_getIOStream_IOStreamClosed_readh(
        FerrisCurlState& state, fh_iostream ss, fh_istream iss )
    {
        const int bsz = 4096;
        char buf[ bsz ];

        iss.read( buf, bsz );
        streamsize actualsz = iss.gcount();
        ss.write( buf, actualsz );

        /* Error writing? */
        if( !ss )
        {
            actualsz = 0;
        }

        return actualsz;
    }
    
    void
    curlFtpContext::priv_getIOStream_IOStreamClosed( fh_istream& ss, std::streamsize tellp )
    {
        LG_CURL_D << "curlFtpContext::priv_getIOStream_IOStreamClosed()"
                  << " path:" << getDirPath()
                  << endl;

//        fh_stringstream ret;
        ss->clear();
        ss->seekg(0);
        ss->clear();
        
        fh_curlstate st = getState( this );
        LG_CURL_D << "About to upload the data " << endl;
        st->setupURL( getMethod(), true );
        st->set( CURLOPT_UPLOAD, 1 );

        streamsize fsz = -1;
//         ss->clear();
//         ss->seekg( 0, ios::end );
//         fsz = ss->tellg();
//         LG_CURL_D << "fsz1:" << fsz << endl;
//         ss->seekg( 0, ios::beg );
//         fsz -= ss->tellg();
//         LG_CURL_D << "fsz2:" << fsz << endl;
//        st->set( CURLOPT_INFILESIZE, fsz );

        LG_CURL_D << "curlFtpContext::priv_getIOStream_IOStreamClosed()"
                  << " Determined size of data at:" << fsz
                  << endl;

        /* Rename the file after the upload is complete */
//         {
//             fh_stringstream ss; ss << "RNFR " << uploadfilename;
//             st->appendFtpCmdPost( tostr(ss) );
//         }
//         {
//             fh_stringstream ss; ss << "RNTO " << realfilename;
//             st->appendFtpCmdPost( tostr(ss) );
//         }

        try
        {
            SigC::Connection con = st->getReadSig().connect(
                SigC::bind(
                    SigC::slot( *this, &curlFtpContext::priv_getIOStream_IOStreamClosed_readh ), ss));
            CurlPerform( st );
            con.disconnect();
        }
        catch( CurlPerformFailed& e )
        {
            if( e.getResultCode() == CURLE_FTP_COULDNT_STOR_FILE )
            {
                fh_stringstream ss;
                ss << " Upload permission denied. path:" << getDirPath()
                   << endl;
                Throw_getIOStreamCloseUpdateFailed( tostr(ss), 0 );
            }
            fh_stringstream ss;
            ss << " Upload not 100% complete for path:" << getDirPath()
               << " determined data size:" << fsz
               << " data actually uploaded:" << st->getBytesReadCount()
               << " e:" << e.what()
               << endl;
            Throw_getIOStreamCloseUpdateFailed( tostr(ss), 0 );
        }

        /*
         * If we are dealing with a string stream then we may not have the end bit
         * set for us. The best we can do is to seek to the end of the stream and
         * check that is is the same tellg() as we had before the seekg.
         */
        {
            streamsize beforeseekg = ss.tellg();
            ss.seekg( 0, ios::end );
            streamsize afterseekg = ss.tellg();

            LG_CURL_D << " beforeseekg:" << beforeseekg
                      << " afterseekg:" << afterseekg
                      << endl;

            if( beforeseekg != afterseekg || ss.fail() )
            {
                fh_stringstream ss;
                ss << " Upload not 100% complete for path:" << getDirPath()
                   << " data actually uploaded:" << st->getBytesReadCount()
                   << " eofbit:" << ss.eof()
                   << " failbit:" << ss.fail()
                   << " beforeseekg:" << beforeseekg
                   << " afterseekg:" << afterseekg
                   << endl;
                Throw_getIOStreamCloseUpdateFailed( tostr(ss), 0 );
            }
        }
        
        if( fsz != -1 )
        {
            if( st->getBytesReadCount() != fsz )
            {
                fh_stringstream ss;
                ss << " Upload not 100% complete for path:" << getDirPath()
                   << " determined data size:" << fsz
                   << " data actually uploaded:" << st->getBytesReadCount()
                   << endl;
                Throw_getIOStreamCloseUpdateFailed( tostr(ss), 0 );
            }
        }
    }
    
    
    fh_iostream
    curlFtpContext::priv_getIOStream( ferris_ios::openmode m )
        throw (FerrisParentNotSetError,
               AttributeNotWritable,
               CanNotGetStream,
               exception)
    {
        LG_CURL_D << "getting IO stream..." << endl;

        fh_stringstream ss;
        fh_iostream iss = fh_stringstream();

        /* dont create a zero byte file */
//         try
//         {
//             iss = getRealStream( m );
//         }
//         catch( CurlPerformFailed& e )
//         {
//             /* they may be making a new file */
//             if( e.getResultCode() == CURLE_FTP_COULDNT_RETR_FILE )
//             {
//             }
//             else throw;
//         }
        
        std::copy( std::istreambuf_iterator<char>(iss),
                   std::istreambuf_iterator<char>(),
                   std::ostreambuf_iterator<char>(ss));
        ss << flush;
        ss->clear();
        ss->seekg(0);
        ss->seekp(0);
        ss->clear();
        
        ss->getCloseSig().connect(SigC::slot( *this, &_Self::priv_getIOStream_IOStreamClosed ));
        return ss;
    }

    void
    curlFtpContext::getRealStream_wrh( FerrisCurlState& state,
                                       fh_iostream& ss,
                                       bool headerSignalFired,
                                       stringset_t* strset,
                                       fh_stringstream r )
    {
        if( !headerSignalFired )
        {
            LG_CURL_D << "firing header event!" << endl;
            headerSignalFired = true;
            getContextEvent_Headers_Received_Sig().emit( this, *strset );
            LG_CURL_D << "done firing header event!" << endl;
        }
        
        std::copy( std::istreambuf_iterator<char>(ss),
                   std::istreambuf_iterator<char>(),
                   std::ostreambuf_iterator<char>(r));
    }

    void
    curlFtpContext::getRealStream_header_wrh( FerrisCurlState* s,
                                              string hdr,
                                              stringset_t* strset )
    {
        fh_stringstream ss;
        ss << hdr;
        string k;
        string v;
        getline( ss, k, ':' );
        getline( ss, v );

        try
        {
            
            if( s->isHTTP() )
            {
                if( k == "Last-Modified" )
                {
                    PrefixTrimmer trimmer;
                    trimmer.push_back( " " );
                    trimmer.push_back( "\t" );
                    v = trimmer( v );
                    
                    struct tm t = Time::ParseTimeString( v, "" );
                    mtime = mktime( &t );
                    LG_CURL_D << "Got header for mtime! remote mtime is:" << mtime << endl;
                    strset->insert( "mtime" );
//                     strset->insert( "mtime-display" );
//                     strset->insert( "mtime-ctime" );
                }
                else if( k == "Content-Length" )
                {
                    LG_CURL_D << "SETTING theSize2 to:" << v << endl;
                    theSize = toType<streamsize>( v );
                    LG_CURL_D << "Got header for sz! remote is:" << theSize << endl;
                    strset->insert( "size" );
//                     strset->insert( "size-human-readable" );
                }
            }
        }
        catch( exception& e )
        {
            LG_CURL_W << "Problem parsing header k:" << k << " v:" << v << endl;
        }
    }
    
    fh_iostream
    curlFtpContext::getRealStream( ferris_ios::openmode m )
        throw (FerrisParentNotSetError,
               CanNotGetStream,
               exception)
    {
        fh_stringstream ret;

        fh_curlstate st = getState( this );
        LG_CURL_D << "About to call st->setupURL() in getRealStream() meth:" << getMethod() << endl;
//         cerr << "About to call st->setupURL() in getRealStream() meth:" << getMethod() << endl;
//         st->print();
//         cerr << "st->serv:" << st->getFtpDetails().serv << endl;
        
        st->setupURL( getMethod(), true );

        LG_CURL_D << "Setting custom CURL option of mtime to v:"
                  << getDownloadIfMTimeSince() << endl;
//        st->set( CURLOPT_TIMEVALUE, getDownloadIfMTimeSince() );
        st->setDownloadIfMTimeSince( getDownloadIfMTimeSince() );

        try
        {
            bool headerSignalFired = false;
            stringset_t strset;
            
            SigC::Connection hcon = st->getHeaderSig().connect(
                SigC::bind( SigC::slot( *this, &curlFtpContext::getRealStream_header_wrh ), &strset ) );
            SigC::Connection wcon = st->getWriteSig().connect(
                SigC::bind(
                    SigC::bind(
                        SigC::bind(
                            SigC::slot( *this, &curlFtpContext::getRealStream_wrh ),
                            ret ),
                        &strset ),
                    headerSignalFired )
                );
            CurlPerform( st );
            hcon.disconnect();
            wcon.disconnect();
        }
        catch( ContentNotModified& e )
        {
            fh_stringstream ss;
            ss << "Remote document not modified, no source for:" << getURL() << endl;
            Throw_ContentNotModified( tostr(ss), 0 );
        }
        catch( exception& e )
        {
            throw;
        }
        

        return ret;
    }
    
    void
    curlFtpContext::priv_createSubContext_writeh( FerrisCurlState& state, fh_iostream& ss )
    {
    }
    
    fh_context
    curlFtpContext::priv_createSubContext( const string& rdn,  fh_context md )
        throw( FerrisCreateSubContextFailed, FerrisCreateSubContextNotSupported )
    {
        const string filename = rdn;
        const string fqfn = appendToPath(getDirPath(),filename);

        LG_CURL_D << "curlFtpContext::createSubContext() rdn :" << rdn << endl;
        LG_CURL_D << "curlFtpContext::createSubContext() path:" << getDirPath() << endl;

        if( getStrAttr( md, CreateObjectType_k, "" ) == CreateObjectType_v_Dir )
        {
            fh_curlstate st = getState( this );
            st->setupURL( getMethod(), false );

            /* We need a MKD command to make the new dir */
            fh_stringstream MKDss;
            MKDss << "MKD " << rdn;
            string MKDCommand = tostr(MKDss);
            st->appendFtpCmdPost( MKDCommand );

            try
            {
                SigC::Connection con = st->getWriteSig().connect(
                    SigC::slot( *this, &curlFtpContext::priv_createSubContext_writeh ));
                CurlPerform( st );
                con.disconnect();
            }
            catch( CurlPerformFailed& e )
            {
                fh_stringstream ss;
                ss << "Error executing a MKD operation. MKDCommand:" << MKDCommand
                   << " path:" << getDirPath()
                   << endl;
                Throw_FerrisCreateSubContextFailed( tostr(ss), this );
            }
        }
        else if( getStrAttr( md, CreateObjectType_k, "" ) == CreateObjectType_v_File )
        {
            try
            {
                curlFtpContext* child = new curlFtpContext( this, rdn );
                fh_context exception_destroyer = child;

                /* This will upload a 0 byte file to the server */
                if( getStrAttr( md, CreateObjectType_k_FileForceCreate, "" ).length() )
                {
                    fh_iostream ios = child->getIOStream();
                }
                addNewChild( child );
            }
            catch( exception& e )
            {
                fh_stringstream ss;
                ss << "Error creating a new file on the ftp server"
                   << " path:" << getDirPath()
                   << " new filename rdn:" << rdn
                   << " e:" << e.what()
                   << endl;
                Throw_FerrisCreateSubContextFailed( tostr(ss), this );
            }
        }
        else
        {
            Throw_FerrisCreateSubContextFailed("curlFtpContext/Unsupported subcontext type.",this);
        }

        return priv_readSubContext( filename, true );
    }

    
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/


    time_t FerrisCurlState::getinfo_FILETIME()
    {
        time_t t = 0;
                
        CURL* c = FerrisCurl::Instance().getCurl();
        CURLcode rc = curl_easy_getinfo( c, CURLINFO_FILETIME, &t );
        if( rc == CURLE_OK )
            return t;
                
        return t;
    }
    
    streamsize
    FerrisCurlState::getinfo_CONTENT_LENGTH_DOWNLOAD()
    {
        double t = 0;
         
        CURL* c = FerrisCurl::Instance().getCurl();
        CURLcode rc = curl_easy_getinfo( c, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &t );
        if( rc == CURLE_OK )
            return streamsize(t);

        LG_CURL_W << "getinfo_CONTENT_LENGTH_DOWNLOAD() failed to get size!" << endl;
        return streamsize(0);
    }

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/


    FerrisCurlClass::FerrisCurlClass()
        :
        c(0)
    {
        FerrisCurlGlobal::Instance();

        c = curl_easy_init();

        if( !c )
        {
            fh_stringstream ss;
            ss << "curl_easy_init() failed";
            Throw_CurlEasyInitFailed( tostr(ss), 0 );
        }
    }

    FerrisCurlClass::~FerrisCurlClass()
    {
        if( c )
        {
            curl_easy_cleanup( c );
        }
    }

    CURL*
    FerrisCurlClass::getCurl()
    {
        return c;
    }
    
    fh_curlstate
    FerrisCurlClass::operator()( curlFtpContext* ctx, fh_curlstate state )
    {
        LG_CURL_D << "FerrisCurlClass::operator() top" << endl;
        state->setup( c );

        char errbuf[CURL_ERROR_SIZE];
        errbuf[0] = 0;
        curl_easy_setopt( c, CURLOPT_ERRORBUFFER, errbuf );
        LG_CURL_D << "FerrisCurlClass::operator() 2" << endl;

        CURLcode r = curl_easy_perform( c );
        LG_CURL_D << "FerrisCurlClass::operator() r:" << r << endl;

        switch(r)
        {
        case CURLE_OK:
            LG_CURL_D << "FerrisCurlClass::operator() CURL_OK" << endl;
            break;
        default:
            fh_stringstream ss;
            ss << "curl_easy_perform() failed:" << errbuf
               << " return code:" << r;
            LG_CURL_D << "E :" << tostr(ss) << endl;
            cerr << tostr(ss) << endl;
            BackTrace();
            Throw_CurlPerformFailed( tostr(ss), 0, r );
        }

        LG_CURL_D << "FerrisCurlClass::operator() isHTTP:" << state->isHTTP() << endl;
        if( state->isHTTP() )
        {
            long htcode;
            curl_easy_getinfo( c, CURLINFO_HTTP_CODE, &htcode );

            LG_CURL_D << "FerrisCurlClass::operator() htcode:" << htcode << endl;
                    
            // not modified 
            if( htcode == 304 )
            {
                fh_stringstream ss;
                Throw_ContentNotModified( tostr(ss), 0 );
            }
        }
                
        LG_CURL_D << "FerrisCurlClass::operator() returning" << endl;
        return state;
    }
    
    
    
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


    extern "C"
    {
        void sig_alarm_cb(int sign)
        {
        }
        
        fh_context Brew( RootContextFactory* rf )
            throw( RootContextCreationFailed )
        {
            try
            {
#ifdef CURL_LEAVES_SIG_ALARM_ON
                struct sigaction newinth;
                newinth.sa_handler = sig_alarm_cb;
                sigemptyset(&newinth.sa_mask);
                newinth.sa_flags   = SA_RESTART;
                if( -1 == sigaction( SIGALRM, &newinth, NULL))
                {
//                    cerr << "ERROR: signal handling is not active. " << endl;
                }
#endif
                
                static curlFtpRootContext* c = 0;
                const string& serv = rf->getInfo( RootContextFactory::SERVERNAME );
                const string& root = rf->getInfo( RootContextFactory::ROOT );
                const string& path = rf->getInfo( RootContextFactory::PATH );

                if( !c )
                {
                    LG_CURL_D << "Making FakeInternalContext(1) " << endl;
                    c = new curlFtpRootContext(0, "/");
            
                    // Bump ref count.
                    static fh_context keeper = c;
                    static fh_context keeper2 = keeper;
                    LG_CURL_D << "Making FakeInternalContext(3) " << endl;
                }

                LG_CURL_D << "Making FakeInternalContext(4) brewing return" << endl;

                fh_context ret = c;

                LG_CURL_D << "curl.brew() root:" << root << " path:" << path << endl;
                
                if( root != "/" )
                {
                    fh_stringstream ss;
                    ss << root << "/" << path;
                    rf->AddInfo( RootContextFactory::PATH, tostr(ss) );
                }

                /*
                 * a hack to make the requested path exist without checking
                 * if it does on the server
                 */
                {
                    fh_stringstream ss;
                    string s;
                    curlFtpContext* ctx = dynamic_cast<curlFtpContext*>(c);

                    LG_CURL_D << "Brew() root:" << root << endl;
                    ss << root << "/" << path;
//                     ss << root;
                    for( bool v=true; getline( ss, s, '/' ); )
                    {
                        if( !s.length() )
                            continue;
                        
                        LG_CURL_D << "for() s:" << s << endl;
                        if(v)
                        {
                            if( !c->isSubContextBound( s ) )
                                break;
                            
                            fh_context tc = c->getSubContext( s );
                            ctx = dynamic_cast<curlFtpContext*>( GetImpl(tc) );
                        }
                        else
                        {
                            curlFtpContext* child = new curlFtpContext( ctx, s );
                            ctx->addNewChild( child );
                            ctx = child;
                        }
                        v=false;
                    }
                }

                LG_CURL_D << "curl.brew() ret:" << ret->getURL() << endl;
                
                return ret;
            }
            catch( FerrisSqlServerNameNotFound& e )
            {
                Throw_RootContextCreationFailed( e.what(), 0 );
            }
            catch( exception& e )
            {
                LG_CURL_D << "Brew() e:" << e.what() << endl;
                Throw_RootContextCreationFailed( e.what(), 0 );
            }
        }
    }


 
};
